// dirutil.cpp
// Revision 11-dec-2003

#include "dirutil.h"

#include <stdexcept>

#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>
#include <sys/types.h>
#include <dirent.h>

namespace {

class AutoCString {
public:
	AutoCString (const char * str) : str (str)
	{ }
	~AutoCString ()
	{
		free (const_cast <char *> (str) );
	}
private:
	const char * str;
};

class AutoDir {
public:
	AutoDir (DIR * dir) : dir (dir)
	{ }
	~AutoDir ()
	{
		closedir (dir);
	}
private:
	DIR * dir;
};

bool has_leading_slash (const std::string & str)
{
	if (str.empty () )
		return false;
	return str [str.size () - 1] == '/';
}

} // namespace

bool dirutil::is_directory (const std::string & str)
{
	struct stat st;
	if (stat (str.c_str (), & st) != 0)
		return false;
	if (! S_ISDIR (st.st_mode) )
		return false;
	return true;
}

std::string dirutil::base_name (const std::string & str)
{
	// basename requires a writable c string.
	char * aux= strdup (str.c_str () );
	if (! aux)
		throw std::bad_alloc ();
	AutoCString aaux (aux);
	return basename (aux);
}

std::string dirutil::dir_name (const std::string & str)
{
	// basename requires a writable c string.
	char * aux= strdup (str.c_str () );
	if (! aux)
		throw std::bad_alloc ();
	AutoCString aaux (aux);
	return dirname (aux);
}

std::string dirutil::extension (const std::string & str)
{
	//TraceFunc tr ("file::extension");

	const std::string file= base_name (str);
	//tr.message (file);
	std::string::size_type l= file.rfind ('.');
	if (l == std::string::npos)
		return std::string ();
	if (l == file.size () )
		return std::string ();
	else return file.substr (l + 1);	
}

std::string dirutil::enter_dir (const std::string & old,
	const std::string & enter)
{
	if (enter == ".")
		return old;
	if (enter == "..")
	{
		std::string base= base_name (old);
		if (base == "/")
			return old;
		if (base != ".." && base != ".")
			return dir_name (old);
	}
	return old + (has_leading_slash (old) ? "" : "/") + enter;
}

void dirutil::directory (const std::string & dirname,
	std::vector <std::string> & v)
{
	v.clear ();
	DIR * dir= opendir (dirname.c_str () );
	if (! dir)
		return;
	AutoDir adir (dir);
	struct dirent * ent;
	while ( (ent= readdir (dir) ) )
		v.push_back (ent->d_name);
}

// End of dirutil.cpp
