// finder.h
// Revision 12-mar-2004

#include <string>

class Finder {
public:
	typedef std::string string;
	typedef std::string::size_type size_type;
	static const size_type npos= string::npos;

	virtual bool backwards () const { return false; }
	virtual size_type find (const string & str)= 0;
	virtual size_type find (const string & str, size_type from) = 0;
	virtual size_type length () const = 0; // Length of last match.
	virtual string replace (const string & str, const string & line);

	virtual ~Finder () { }
};

class StringFinder : public Finder {
public:
	StringFinder (const string & searched) :
		searched (searched),
		l (searched.size () )
	{ }
	size_type length () const { return l; }
	virtual bool adjustfrom (const string & str, size_type & from);
protected:
	const string searched;
private:
	const size_type l;
};

class NormalFinder : public StringFinder {
public:
	NormalFinder (const string & searched) :
		StringFinder (searched)
	{ }
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
};

class BackFinder : public StringFinder {
public:
	BackFinder (const string & searched) :
		StringFinder (searched)
	{ }
	bool backwards () const { return true; }
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
};
class NocaseFinder : public StringFinder {
public:
	NocaseFinder (const string & searched) :
		StringFinder (searched)
	{ }
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
};

class BackNocaseFinder : public StringFinder {
public:
	BackNocaseFinder (const string & searched) :
		StringFinder (searched)
	{ }
	bool backwards () const { return true; }
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
private:
	bool adjustfrom (const string & str, size_type & from);
};

class WordBaseFinder : public StringFinder {
public:
	WordBaseFinder (const string & searched);

	size_type find (const string & str);
	size_type find (const string & str, size_type from) ;
protected:
	bool checkword (const string & str, size_type r);
	const size_type ssize;
	const bool alnumfirst;
	const bool alnumlast;

	virtual size_type do_find (const string & str, size_type from)= 0;
};

class WordFinder : public WordBaseFinder {
public:
	WordFinder (const string & searched) :
		WordBaseFinder (searched)
	{ }
private:
	virtual size_type do_find (const string & str, size_type from);
};

class WordBackFinder : public WordBaseFinder {
public:
	WordBackFinder (const string & searched) :
		WordBaseFinder (searched)
	{ }
	bool backwards () const { return true; }
private:
	virtual size_type do_find (const string & str, size_type from);
};

class WordNocaseFinder : public WordBaseFinder {
public:
	WordNocaseFinder (const string & searched) :
		WordBaseFinder (searched)
	{ }
private:
	virtual size_type do_find (const string & str, size_type from);
};

class WordBackNocaseFinder : public WordBaseFinder {
public:
	WordBackNocaseFinder (const string & searched) :
		WordBaseFinder (searched)
	{ }
	bool backwards () const { return true; }
private:
	bool adjustfrom (const string & str, size_type & from);
	virtual size_type do_find (const string & str, size_type from);
};

class RegexBaseFinder : public Finder {
public:
	RegexBaseFinder (const string & searched, bool nocase);
	~RegexBaseFinder ();

	virtual size_type find (const string & str)= 0;
	virtual size_type find (const string & str, size_type from)= 0;
	size_type length () const;
protected:
	class Internal;
	Internal * pin;
private:
	string replace (const string & str, const string & line);
};

class RegexFinder : public RegexBaseFinder {
public:
	RegexFinder (const string & searched, bool nocase);
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
};

class RegexBackFinder : public RegexBaseFinder {
public:
	RegexBackFinder (const string & searched, bool nocase);
	bool backwards () const { return true; }
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
};

// End of finder.h
