// menu.cpp
// Revision 5-dec-2003

#include "menu.h"
#include "util.h"
#include "trace.h"

#include <stdexcept>
#include <algorithm>

#include "qtreassert.h"

// *********************************************************
// *                      Option                           *
// *********************************************************

size_t Option::numopts= 0;

#if 0
Option::Option (std::string ntext, IdMenu nid) :
	str (parsestr (ntext) ),
	id (nid),
	owner (false),
	to (OptionIsCom)
{
	++numopts;
}
#endif

Option::Option (std::string ntext, Menu * nmenu, MenuOwnership ownership) :
	str (parsestr (ntext) ),
	//menu (new Menu (* nmenu) ),
	menu (nmenu),
	owner (ownership == MenuOwner),
	to (OptionIsSub)
{
	++numopts;
}

Option::Option (std::string ntext, Action * naction, MenuOwnership ownership) :
	str (parsestr (ntext) ),
	action (naction),
	owner (ownership == MenuOwner),
	to (OptionIsAction)
{
	++numopts;
}



#if 0
Option::Option (const Option & opt) :
	str (opt.str),
	id (opt.id),
	to (opt.to),
	position (opt.position),
	abrev (opt.abrev)
{
	++numopts;
	if (to == OptionIsSub)
		menu= new Menu (* opt.menu);
}

Option::Option & Option::operator = (const Option & opt)
{
	str= opt.str;
	id= opt.id;
	Menu * nmenu= 0;
	if (opt.to == OptionIsSub)
		nmenu= new Menu (* opt.menu);
	if (to == OptionIsSub)
		delete menu;
	menu= nmenu;
	to= opt.to;
	position= opt.position;
	abrev= opt.abrev;
	return * this;
}
#endif

Option::~Option ()
{
	TraceFunc tr ("Option::~Option");

	if (to == OptionIsSub && owner)
		delete menu;
	ASSERT (numopts > 0);
	--numopts;
}

std::string Option::parsestr (const std::string & str)
{
	using std::string;
	abrev= '\0';
	string::size_type pos= str.find ('&');
	if (pos != string::npos)
	{
		if (pos == str.size () -1)
			throw std::runtime_error ("Bad menu");
		abrev= str [pos + 1];
		string s= str;
		s.erase (pos, 1);
		return s;
	}
	return str;
}

// *********************************************************
// *                      Menu                             *
// *********************************************************

Menu::Menu () :
	pot (0),
	width (0),
	inipos (0),
	numopts (0),
	active (NO_SEL)
{
	TraceFunc tr ("Menu::Menu");
	tr.message ("width= " + util::to_string (width) );
}

Menu::~Menu ()
{
	TraceFunc tr ("Menu::~Menu");

	for (size_t i= 0, l= options.size (); i < l; ++i)
		if (options [i].owner)
			delete options [i].pop;
}

void Menu::addOption (Option * pop, MenuOwnership ownership)
{
	options.push_back (MenuOption (pop, ownership == MenuOwner) );
}

void Menu::insertOption (Option * pop, MenuOwnership ownership,
	size_t position)
{
	if (position > options.size () )
		addOption (pop, ownership);
	else
		options.insert (options.begin () + position,
			MenuOption (pop, ownership == MenuOwner) );
}

Menu * Menu::getSubMenu (const std::string & name)
{
	for (size_t i= 0, l= options.size (); i < l; ++i)
		if (options [i].type () == Option::OptionIsSub &&
			options [i].text () == name)
		{
			return & options [i].submenu ();
		}
	return NULL;
}

void Menu::calcpositions ()
{
	width= pot->columns ();
	clear= std::string (width, ' ');

	numopts= options.size ();
	size_t position= 0;
	for (size_t i= 0; i < numopts; ++i)
	{
		options [i].pop->pos (position);
		position+= options [i].pop->text ().size () + 1;
	}
}

void Menu::putopt (size_t nopt)
{
	size_t pos= options [nopt].pop->pos ();
	const std::string & text= options [nopt].pop->text ();
	* pot << at (0, pos - inipos) <<
		' ' << text << ' ' << std::flush;
}

void Menu::putall ()
{
	for (size_t i= 0; i < numopts; ++i)
	{
		const Option & opt= * options [i].pop;
		if (opt.pos () >= inipos &&
			//opt.pos () + opt.text ().size () <= inipos + width)
			opt.pos () < inipos + width);
		{
			if (i == active)
				* pot << inverse (false);
			putopt (i);
			if (i == active)
				* pot << inverse (true);
		}
	}
}

void Menu::setactive (size_t newactive)
{
	TraceFunc tr ("Menu::setactive");

	if (newactive == active)
		return;
	if (active != NO_SEL)
	{
		putopt (active);
	}
	active= newactive;
	if (newactive != NO_SEL)
	{
		size_t pos= options [active].pop->pos ();
		const std::string & text= options [active].pop->text ();
		tr.message ("endpos= " + util::to_string
			(pos + text.size () ) );
		if (pos < inipos)
		{
			inipos= pos;
			//putall ();
			show ();
		}
		else if (pos + text.size () >= inipos + width - 2)
		{
			size_t aux= pos + text.size () - width + 2;
			size_t n= 0;
			for ( ; options [n].pop->pos () < aux; ++n)
				continue;
			inipos= options [n].pop->pos ();
			//putall ();
			show ();
		}
		else
		{
			* pot << inverse (false);
			putopt (active);
			* pot << inverse (true);
		}
	}
}

void Menu::show ()
{
	TraceFunc ("Menu::show");

	* pot << at (0, 0) << inverse (true) << clear;
	//for (size_t i= 0; i < options.size (); ++i)
	//	putopt (i);
	putall ();
}

void Menu::show (otermstream & ots)
{
	TraceFunc ("Menu::show");

	pot= & ots;
	calcpositions ();
	show ();
}

//bool Menu::select (IdMenu * selected, otermstream & ots)
Action * Menu::select (otermstream & ots)
{
	TraceFunc tr ("Menu::select");

	using namespace key;

	if (options.empty () )
		return NULL;
	pot= & ots;
	show (ots);
	//ASSERT (selected != 0);
	//size_t active= 0;
	setactive (0);
	bool fSelecting= true;
	bool fSelected= false;
	Action * result= NULL;
	do
	{
		int c;
		switch ( (c= pot->getkey () ) )
		{
		case '\r': case '\n':
			if ( (result= activate (active, ots) ) != 0)
			{
				fSelected= true;
				fSelecting= false;
			}
			//redraw (newactive);
			show ();
			break;
		case Esc:
		case F8:
			fSelecting= false;
			break;
		case Left:
			setactive ( (active + numopts - 1) % numopts);
			break;
		case Right:
			setactive ( (active + 1) % numopts);
			break;
		default:
			if (c > 0xFF)
				break;
			c= toupper (c);
			if (options [active].pop->charabrev () != c)
			{
				size_t i;
				for (i= (active + 1) % numopts;
					i!= active; i= (i + 1) % numopts)
					if (options [i].pop->charabrev () == c)
						break;
				if (i == active)
					break;
				setactive (i);
			}
			if ( (result= activate (active, ots) ) != 0)
			{
				fSelected= true;
				fSelecting= false;
			}
			//redraw (newactive);
			show ();
			//prev= NO_SEL;
		} // switch
	} while (fSelecting);
	* pot << inverse (true);
	//putopt (newactive);
	setactive (NO_SEL);
	inipos= 0;
	show ();
	//return fSelected;
	if (fSelected)
	{
		ASSERT (result != 0);
		return result;
	}
	else
		return 0;
}

//bool Menu::activate (size_t n, IdMenu * selected, otermstream & ots)
Action * Menu::activate (size_t n, otermstream & ots)
{
	if (options [n].pop->type () == Option::OptionIsSub)
	{
		Menu & submenu= options [n].pop->submenu ();
		#if 0
		if (submenu.select (selected, ots) )
			return true;
		else return false;
		#else
		return submenu.select (ots);
		#endif
	}
	else
	{
		#if 0
		* selected= options [n].pop->getId ();
		return true;
		#else
		return options [n].pop->getAction ();
		#endif
	}
}

// *********************************************************
// *                    MapKeyAction                       *
// *********************************************************

MapKeyAction::~MapKeyAction ()
{
	for_each (m.begin (), m.end (), Destroy () );
}

// Fin de menu.cpp
