// menu.h
// Revision 1-dec-2003

#include <string>
#include <vector>
#include <map>
#include "termstream.h"

//typedef unsigned int IdMenu;

enum MenuOwnership { MenuOwner, MenuNotOwner };

class Action {
public:
	virtual void doit ()= 0;
	virtual ~Action ()
	{
	}
};

class Menu;

class Option {
public:
	typedef size_t pos_t;
	typedef enum { OptionIsCom, OptionIsSub, OptionIsAction } TypeOption;
	//Option (std::string ntext, IdMenu nid);
	Option (std::string ntext, Menu * nmenu, MenuOwnership ownership);
	Option (std::string ntext, Action * naction, MenuOwnership ownership);
	~Option ();
	const std::string & text () const { return str; }
	//IdMenu getId () { return id; }
	Action * getAction () { return action; }
	void pos (pos_t npos) { position= npos; }
	pos_t pos () const { return position; }
	TypeOption type () const { return to; }
	Menu & submenu () { return * menu; }
	char charabrev () const { return abrev; }
private:
	Option (const Option & opt); // Forbidden
	Option & operator= (const Option & opt); // Forbidden
	std::string str;
	//IdMenu id;
	union {
		Menu * menu;
		Action * action;
	};
	bool owner;
	TypeOption to;
	pos_t position;
	char abrev;
	std::string parsestr (const std::string & str);
	static size_t numopts;
};

class Menu {
public:
	Menu ();
	~Menu ();
	void addOption (Option * pop, MenuOwnership ownership);
	void insertOption (Option * pop, MenuOwnership ownership,
		size_t position);
	Menu * getSubMenu (const std::string & name);
	void show ();
	void show (otermstream & ots);
	//bool select (IdMenu * selected, otermstream & ots);
	Action * select (otermstream & ots);
private:
	Menu (const Menu &); // Forbidden
	void operator= (const Menu &); // Forbidden
	otermstream * pot;
	struct MenuOption {
		Option * pop;
		bool owner;
		MenuOption (Option * pop, bool owner) :
			pop (pop), owner (owner)
		{ }
		Option::TypeOption type () const { return pop->type (); }
		std::string text () const { return pop->text (); }
		Menu & submenu () { return pop->submenu (); }
	};
	std::vector <MenuOption> options;
	size_t width;
	std::string clear;
	size_t inipos;
	size_t numopts;
	static const size_t NO_SEL= size_t (-1);
	size_t active;

	void calcpositions ();
	void putopt (size_t nopt);
	void putall ();
	void setactive (size_t newactive);
	//bool activate (size_t n, IdMenu * selected, otermstream & ots);
	Action * activate (size_t n, otermstream & ots);
};

class MapKeyAction {
	typedef std::map <int, Action *> mapka;
	mapka m;
	struct Destroy {
		void operator () (mapka::value_type & v)
		{
			delete v.second;
		}
	};
public:
	virtual ~MapKeyAction ();
	void addKeyAction (int key, Action * action)
	{
		mapka::iterator it= m.find (key);
		if (it != m.end () )
			Destroy () (* it);
		m [key]= action;
	}
	Action * getKeyAction (int key)
	{
		mapka::iterator it= m.find (key);
		if (it != m.end () )
			return it->second;
		else
			return 0;
	}
};

// Fin de menu.h
