/*
	qperl.c
	Revision 7-dec-2003
*/

#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "qperl.h"

#define QTRE_PREFIX "qtre::"
#define MAXPATH 128

static const char qtre_prefix []= QTRE_PREFIX;
static const char qtre_prefix_default []= QTRE_PREFIX "default::";
static const char strGlobalPl []= "/etc/qtre.pl";
static const char strPl []= ".qtre.pl";

#define DIM(a) (sizeof (a) / sizeof (*a) )

void * gdata= NULL;

/* Flag to control if initialisation has been ended
with the function end_init. */
int init_ended= 0;

void errormessage (const char * title, const char * message)
{
	if (gdata)
	{
		qpDialogShowTextParam param;
		param.title= title;
		param.str= message;
		param.height= 0;
		param.width= 60;
		qpCallback (gdata, qpDialogShowText, 0, & param);
	}
	else
		fprintf (stderr, "%s\n%s\n", title, message);
}

void too_much (const char * function)
{
	croak ("Too much parameters to qtre::%s", function);
}

void require_argument (const char * function, int n)
{
	if (n == 1)
		croak ("qtre::%s needs 1 parameter", function);
	else
		croak ("qtre::%s needs %i parameters", function, n);
}

void no_argument (const char * function)
{
	croak ("qtre::%s does not need parameters", function);
}

void no_memory (const char * function)
{
	croak ("memory full in qtre::%s", function);
}

static const char help_mode []=
"Pendiente de escribir la ayuda";

static XS (mode)
{
	static const char function []= "mode";
	int mode;
	dXSARGS;
	if (items != 1)
	{
		require_argument (function, 1);
		XSRETURN_UNDEF;
	}
	mode= SvIV (ST (0) );
	if (qpCallback (gdata, qpMode, mode, NULL) == 0)
	{
		croak ("qtre::mode: Invalid mode %i", mode);
		XSRETURN_UNDEF;
	}
	XSRETURN_YES;
}

static const char help_empty []=
"Pendiente de escribir la ayuda";

static XS (empty)
{
	static const char function []= "empty";
	int result;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	result= qpCallback (gdata, qpIsEmpty, 0, NULL);
	XSRETURN_IV (result);
}

static const char help_modified []=
"Pendiente de escribir la ayuda";

static XS (modified)
{
	static const char function []= "modified";
	int result;
	dXSARGS;
	switch (items)
	{
	case 0:
		result= qpCallback (gdata, qpGetModified, 0, NULL);
		break;
	case 1:
		result= qpCallback
			(gdata, qpSetModified, SvIV (ST (0) ), NULL);
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	XSRETURN_IV (result);
}

static const char help_up []=
"Pendiente de escribir la ayuda";

static XS (up)
{
	static const char function []= "up";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		if (! qpCallback (gdata, qpUp, 0, NULL) )
			XSRETURN_UNDEF;
	XSRETURN_YES;
}

static const char help_down []=
"Pendiente de escribir la ayuda";

static XS (down)
{
	static const char function []= "down";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		if (! qpCallback (gdata, qpDown, 0, NULL) )
			XSRETURN_UNDEF;
	XSRETURN_YES;
}

static const char help_left []=
"Pendiente de escribir la ayuda";

static XS (left)
{
	static const char function []= "left";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		if (! qpCallback (gdata, qpLeft, 0, NULL) )
			XSRETURN_UNDEF;
	XSRETURN_YES;
}

static const char help_right []=
"Pendiente de escribir la ayuda";

static XS (right)
{
	static const char function []= "right";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		if (! qpCallback (gdata, qpRight, 0, NULL) )
			XSRETURN_UNDEF;
	XSRETURN_YES;
}

static const char help_home []=
"Pendiente de escribir la ayuda";

static XS (home)
{
	static const char function []= "home";
	dXSARGS;
	if (items != 0)
	{
		too_much (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpHome, 0, NULL);
	XSRETURN_YES;
}

static const char help_end []=
"Pendiente de escribir la ayuda";

static XS (end)
{
	static const char function []= "end";
	dXSARGS;
	if (items != 0)
	{
		too_much (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpEnd, 0, NULL);
	XSRETURN_YES;
}

static const char help_pageup []=
"Pendiente de escribir la ayuda";

static XS (pageup)
{
	static const char function []= "pageup";
	dXSARGS;
	if (items != 0)
	{
		too_much (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpPageUp, 0, NULL);
	XSRETURN_YES;
}

static const char help_pagedown []=
"Pendiente de escribir la ayuda";

static XS (pagedown)
{
	static const char function []= "pagedown";
	dXSARGS;
	if (items != 0)
	{
		too_much (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpPageDown, 0, NULL);
	XSRETURN_YES;
}

static const char help_goto_line []=
"Pendiente de escribir la ayuda";

static XS (goto_line)
{
	static const char function []= "goto_line";
	int line;
	int n;
	dXSARGS;
	if (items != 1)
	{
		require_argument (function, 1);
		XSRETURN_UNDEF;
	}
	line= SvIV (ST (0) );
	n= qpCallback (gdata, qpGotoLine, line, NULL);
	XSRETURN_IV (n);
}

static const char help_get_char []=
"Pendiente de escribir la ayuda";

static XS (get_char)
{
	static const char function []= "get_char";
	char c;
	SV * sv;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	c= qpCallback (gdata, qpGetChar, 0, NULL);
	if (c != 0)
		sv= sv_2mortal (newSVpv (&c, 1) );
	else sv= & PL_sv_undef;
	/*
	XPUSHs (sv);
	XSRETURN (1);
	*/
	/*
	EXTEND (SP, 1);
	PUSHs (sv);
	PUTBACK;
	XSRETURN (1);
	*/
	ST (0)= sv;
	XSRETURN (1);
}

static const char help_get_word []=
"Pendiente de escribir la ayuda";

static XS (get_word)
{
	static const char function []= "get_word";
	qpStringParam param;
	SV * sv;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	if (! qpCallback (gdata, qpGetWord, 0, & param) )
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	sv= sv_2mortal (newSVpv (param.str, param.len) );
	qpCallback (gdata, qpFreeString, 0, & param);	
	ST (0)= sv;
	PUTBACK;
	XSRETURN (1);
}

static const char help_delete_char []=
"Pendiente de escribir la ayuda";

static XS (delete_char)
{
	static const char function []= "delete_char";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		qpCallback (gdata, qpDeleteChar, 0, NULL);
	XSRETURN_YES;
}

static const char help_backspace []=
"Pendiente de escribir la ayuda";

static XS (backspace)
{
	static const char function []= "backspace";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		qpCallback (gdata, qpBackspace, 0, NULL);
	XSRETURN_YES;
}

static const char help_delete_line []=
"Pendiente de escribir la ayuda";

static XS (delete_line)
{
	static const char function []= "delete_line";
	int i, count;
	dXSARGS;
	switch (items)
	{
	case 0:
		count= 1;
		break;
	case 1:
		count= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	for (i= 0; i < count; ++i)
		qpCallback (gdata, qpDeleteLine, 0, NULL);
	XSRETURN_YES;
}

static const char help_replace_line []=
"Pendiente de escribir la ayuda";

static XS (replace_line)
{
	static const char function []= "replace_line";
	char *str;
	int line;
	STRLEN len;
	dXSARGS;
	/*line= qpCallback (gdata, qpLine, 0, NULL);*/
	switch (items)
	{
	case 0:
		str= "";
		break;
	#if 0
	case 2:
		line= SvIV (ST (1) );
	case 1:
		str= SvPV (ST (0), len);
		break;
	#else
	case 1:
		line= qpCallback (gdata, qpLine, 0, NULL);
		break;
	case 2:
		line= SvIV (ST (1) );
		break;
	#endif
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	str= SvPV (ST (0), len);
	qpCallback (gdata, qpReplaceLine, line, str);
	XSRETURN_YES;
}

static const char help_insert_char []=
"Pendiente de escribir la ayuda";

static XS (insert_char)
{
	static const char function []= "insert_char";
	char *str;
	int n;
	STRLEN len;
	dXSARGS;
	if (items != 1)
	{
		require_argument (function, 1);
		XSRETURN_UNDEF;
	}
	if (SvIOKp (ST (0) ) )
		n= SvIV (ST (0) );
	else
	{
		str= SvPV (ST (0), len);
		if (len != 1)
		{
			croak ("qtre: Invalid parameter to insert_char");
			XSRETURN_UNDEF;
		}
		n= str [0];
	}
	qpCallback (gdata, qpInsertChar, n, NULL);
	XSRETURN_YES;
}

static const char help_insert []=
"Pendiente de escribir la ayuda";

static XS (insert)
{
	static const char function []= "insert";
	qpStringParam param;
	/*char *str;*/
	STRLEN len;
	dXSARGS;
	if (items != 1)
	{
		require_argument (function, 1);
		XSRETURN_UNDEF;
	}
	param.str= SvPV (ST (0), len);
	param.len= len;
	/*qpCallback (gdata, qpInsert, 0, str);*/
	qpCallback (gdata, qpInsert, 0, & param);
	XSRETURN_YES;
}

static const char help_get_beginsel []=
"Pendiente de escribir la ayuda";

static XS (get_beginsel)
{
	static const char function []= "get_beginsel";
	qpSelectionLimitsParam param;
	dXSARGS;
	/*if (items != 0)*/
	if (items > 1) /* Una prueba */
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	if (qpCallback (gdata, qpSelectionLimits, 0, & param) == 0)
		XSRETURN_UNDEF;
	#if 0
	XPUSHs (sv_2mortal (newSViv (param.beginline) ) );
	XPUSHs (sv_2mortal (newSViv (param.beginpos) ) );
	XSRETURN (2);
	#else
	EXTEND (SP, 2);

	#if 0
	PUSHs (sv_2mortal (newSViv (param.beginline) ) );
	PUSHs (sv_2mortal (newSViv (param.beginpos) ) );
	PUTBACK;
	#else
	ST (0)= sv_2mortal (newSViv (param.beginline) );
	ST (1)= sv_2mortal (newSViv (param.beginpos) );
	#endif

	XSRETURN (2);
	#endif
}

static const char help_get_endsel []=
"Pendiente de escribir la ayuda";

static XS (get_endsel)
{
	static const char function []= "get_endsel";
	qpSelectionLimitsParam param;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	if (qpCallback (gdata, qpSelectionLimits, 0, & param) == 0)
		XSRETURN_UNDEF;
	#if 0
	XPUSHs (sv_2mortal (newSViv (param.endline) ) );
	XPUSHs (sv_2mortal (newSViv (param.endpos) ) );
	XSRETURN (2);
	#else
	EXTEND (SP, 2);

	#if 0
	PUSHs (sv_2mortal (newSViv (param.endline) ) );
	PUSHs (sv_2mortal (newSViv (param.endpos) ) );
	PUTBACK;
	#else
	ST (0)= sv_2mortal (newSViv (param.endline) );
	ST (1)= sv_2mortal (newSViv (param.endpos) );
	#endif

	XSRETURN (2);
	#endif
}

static const char help_get_selection []=
"Pendiente de escribir la ayuda";

static XS (get_selection)
{
	static const char function []= "get_selection";
	int l;
	char * str;
	qpStringParam param;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	#if 0
	l= qpCallback (gdata, qpLengthSelection, 0, NULL);
	str= malloc (l + 1);
	if (str == NULL)
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpGetSelection, 0, str);
	ST(0)= sv_2mortal (newSVpv (str, l) );
	free (str);
	#else
	l= qpCallback (gdata, qpGetSelection, 0, & param);
	if (l == 0)
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	ST(0)= sv_2mortal (newSVpv (param.str, (int) param.len) );
	//free (param.str);
	qpCallback (gdata, qpFreeString, 0, & param);
	#endif
	XSRETURN (1);
}

static const char help_get_clipboard []=
"Pendiente de escribir la ayuda";

static XS (get_clipboard)
{
	static const char function []= "get_clipboard";
	int l;
	char * str;
	qpStringParam param;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	#if 0
	l= qpCallback (gdata, qpLengthClipboard, 0, NULL);
	str= malloc (l + 1);
	if (str == NULL)
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpGetClipboard, 0, str);
	ST(0)= sv_2mortal (newSVpv (str, l) );
	free (str);
	#else
	l= qpCallback (gdata, qpGetClipboard, 0, & param);
	if (l == 0)
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	ST(0)= sv_2mortal (newSVpv (param.str, (int) param.len) );
	//free (param.str);
	qpCallback (gdata, qpFreeString, 0, & param);
	#endif
	XSRETURN (1);
}

static const char help_set_clipboard []=
"Pendiente de escribir la ayuda";

static XS (set_clipboard)
{
	static const char function []= "set_clipboard";
	char * str;
	STRLEN len;
	dXSARGS;
	switch (items)
	{
	case 0:
		str= "";
		break;
	case 1:
		str= SvPV (ST (0), len);
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpSetClipboard, 0, str);
	XSRETURN_YES;
}

static const char help_line []=
"Pendiente de escribir la ayuda";

static XS (line)
{
	static const char function []= "line";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpLine, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_num_lines []=
"Pendiente de escribir la ayuda";

static XS (num_lines)
{
	static const char function []= "num_lines";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpNumLines, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_char_line []=
"Pendiente de escribir la ayuda";

static XS (char_line)
{
	static const char function []= "char_line";
	int n;
	int nline;
	dXSARGS;
	switch (items)
	{
	case 0:
		nline= qpCallback (gdata, qpLine, 0, NULL);
		break;
	case 1:
		nline= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpCharLine, nline, NULL);
	XSRETURN_IV (n);
}

static const char help_get_line []=
"Pendiente de escribir la ayuda";

static XS (get_line)
{
	static const char function []= "get_line";
	qpStringParam param;
	int n;
	int nline;
	char * str;
	SV * sv;
	dXSARGS;
	switch (items)
	{
	case 0:
		nline= qpCallback (gdata, qpLine, 0, NULL);
		break;
	case 1:
		nline= SvIV (ST (0) );
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	#if 0
	n= qpCallback (gdata, qpCharLine, nline, NULL);
	str=  malloc (n + 1);
	if (str == NULL)
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpGetLine, nline, str);

	/*ST(0)= sv_2mortal (newSVpv (str, 0) );*/

	EXTEND (SP, 1);
	PUSHs (sv_2mortal (newSVpv (str, 0) ) );

	free (str);
	/*XSRETURN (1);*/
	PUTBACK;
	return;
	/*XSRETURN (1);*/
	#else
	if (! qpCallback (gdata, qpGetLine, nline, & param) )
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	/*
	sv= sv_2mortal (newSVpv (str, 0) );
	free (str);
	*/
	sv= sv_2mortal (newSVpv (param.str, param.len) );
	qpCallback (gdata, qpFreeString, 0, & param);
	/*if (items == 0)*/
	if (0)
	{
		EXTEND (SP, 1);
		PUSHs (sv);
	}
	else
		ST (0)= sv;
	PUTBACK;
	XSRETURN (1);
	#endif
}

static const char help_get_pos []=
"Pendiente de escribir la ayuda";

static XS (get_pos)
{
	static const char function []= "get_pos";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpGetPos, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_get_col []=
"Pendiente de escribir la ayuda";

static XS (get_col)
{
	static const char function []= "get_col";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpGetCol, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_get_fileext []=
"Pendiente de escribir la ayuda";

static XS (get_fileext)
{
	static const char function []= "get_fileext";
	dXSARGS;
	char buf [MAXPATH];
	qpCallback (gdata, qpGetFileExt, MAXPATH, buf);
	XSRETURN_PV (buf);
}

static const char help_get_filename []=
"Pendiente de escribir la ayuda";

static XS (get_filename)
{
	static const char function []= "get_filename";
	dXSARGS;
	char buf [MAXPATH];
	qpCallback (gdata, qpGetFileName, MAXPATH, buf);
	XSRETURN_PV (buf);
}

static const char help_status []=
"Pendiente de escribir la ayuda";

static XS (status)
{
	static const char function []= "status";
	char *str;
	STRLEN len;
	dXSARGS;
	if (items != 1)
	{
		require_argument (function, 1);
		XSRETURN_UNDEF;
	}
	str= SvPV (ST (0), len);
	qpCallback (gdata, qpStatus, 0, str);
	XSRETURN_YES;
}

static const char help_status_question []=
"Pendiente de escribir la ayuda";

static XS (status_question)
{
	static const char function []= "status_question";
	char * question;
	char * values;
	char r;
	qpStatusQuestionParam param;
	STRLEN len;
	dXSARGS;
	if (items != 2)
	{
		require_argument (function, 2);
		XSRETURN_UNDEF;
	}
	param.question= SvPV (ST (0), len);
	param.values= SvPV (ST (1), len);
	r= qpCallback (gdata, qpStatusQuestion, 0, & param);
	ST (0)= sv_2mortal (newSVpv (& r, 1) );
	XSRETURN (1);
}

static const char help_dialog_get_string []=
"Pendiente de escribir la ayuda";

static XS (dialog_get_string)
{
	static const char function []= "dialog_get_string";
	qpDialogGetStringParam param;
	int r;
	STRLEN len;
	dXSARGS;
	if (items < 4)
	{
		require_argument (function, 4);
		XSRETURN_UNDEF;
	}
	param.title= SvPV (ST (0), len);
	param.prompt= SvPV (ST (1), len);
	param.width= SvIV (ST (2) );
	if (items >= 4)
		param.maxlength= SvIV (ST (3) );
	else
		param.maxlength= param.width;
	param.str=  malloc (param.maxlength + 1);
	if (param.str == NULL)
	{
		no_memory (function);
		XSRETURN_UNDEF;
	}
	if (items >= 5)
	{
		char * aux;
		if (items > 5)
		{
			too_much (function);
			XSRETURN_UNDEF;
		}
		aux= SvPV (ST (4), len);
		strncpy (param.str, aux, param.maxlength);
		param.str [param.maxlength]= '\0';
	}
	else
		param.str [0]= '\0';
	r= qpCallback (gdata, qpDialogGetString, 0, & param);
	if (r)
		ST(0)= sv_2mortal (newSVpv (param.str, 0) );
	free (param.str);
	if (r)
		XSRETURN (1);
	else
		XSRETURN_UNDEF;
}

static const char help_dialog_show_text []=
"Pendiente de escribir la ayuda";

static XS (dialog_show_text)
{
	static const char function []= "dialog_show_text";
	qpDialogShowTextParam param;
	STRLEN len;
	dXSARGS;
	if (items < 2)
	{
		require_argument (function, 2);
		XSRETURN_UNDEF;
	}
	param.title= SvPV (ST (0), len);
	param.str= SvPV (ST (1), len);
	param.height= 0;
	param.width= 0;
	if (items >= 3)
	{
		param.height= SvIV (ST (2) );
		if (items >= 4)
		{
			param.width= SvIV (ST (3) );
			if (items >= 5)
			{
				too_much (function);
				XSRETURN_UNDEF;
			}
		}
	}
	qpCallback (gdata, qpDialogShowText, 0, & param);
	XSRETURN_YES;
}

static const char help_define_control []=
"Pendiente de escribir la ayuda";

static XS (define_control)
{
	static const char function []= "define_control";
	qpDefineControlParam param;
	char * str;
	STRLEN len;
	dXSARGS;
	if (items != 2)
	{
		require_argument (function, 2);
		XSRETURN_UNDEF;
	}
	str= SvPV (ST (0), len);
	if (len != 1)
		XSRETURN_UNDEF;
	param.key= str [0];
	param.function= SvPV (ST (1), len);
	qpCallback (gdata, qpDefineControl, 0, & param);
	XSRETURN_YES;
}

static const char help_redraw []=
"Pendiente de escribir la ayuda";

static XS (redraw)
{
	static const char function []= "redraw";
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpRedraw, 0, 0);
	XSRETURN_YES;
}

static const char help_quiet []=
"Pendiente de escribir la ayuda";

static XS (quiet)
{
	static const char function []= "quiet";
	int mode;
	dXSARGS;
	switch (items)
	{
	case 0:
		mode= 1;
		break;
	case 1:
		mode= SvIV (ST (0) );
		break;
	default:
		too_much (function);
	}
	qpCallback (gdata, qpQuiet, mode, 0);
	XSRETURN_YES;
}

static const char help_toggle_select []=
"Pendiente de escribir la ayuda";

static XS (toggle_select)
{
	static const char function []= "toggle_select";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpSelect, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_end_select []=
"Pendiente de escribir la ayuda";

static XS (end_select)
{
	static const char function []= "end_select";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpEndSelect, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_select_all []=
"Pendiente de escribir la ayuda";

static XS (select_all)
{
	static const char function []= "select_all";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpSelectAll, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_cut []=
"Pendiente de escribir la ayuda";

static XS (cut)
{
	static const char function []= "cut";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpCut, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_copy []=
"Pendiente de escribir la ayuda";

static XS (copy)
{
	static const char function []= "copy";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpCopy, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_paste []=
"Pendiente de escribir la ayuda";

static XS (paste)
{
	static const char function []= "paste";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpPaste, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_get_key []=
"Pendiente de escribir la ayuda";

static XS (get_key)
{
	static const char function []= "get_key";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpGetKey, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_autoindent []=
"Pendiente de escribir la ayuda";

static XS (autoindent)
{
	static const char function []= "autoindent";
	int n;
	dXSARGS;
	switch (items)
	{
	case 0:
		n= qpCallback (gdata, qpGetAutoindent, 0, NULL);
		break;
	case 1:
		n= qpCallback (gdata, qpSetAutoindent, SvIV (ST (0) ), NULL);
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	XSRETURN_IV (n);
}

static const char help_autocr []=
"Pendiente de escribir la ayuda";

static XS (autocr)
{
	static const char function []= "autocr";
	int n;
	dXSARGS;
	switch (items)
	{
	case 0:
		n= qpCallback (gdata, qpGetAutocr, 0, NULL);
		break;
	case 1:
		n= qpCallback (gdata, qpSetAutocr, SvIV (ST (0) ), NULL);
		break;
	default:
		too_much (function);
		XSRETURN_UNDEF;
	}
	XSRETURN_IV (n);
}

static const char help_menu_add []=
"Pendiente de escribir la ayuda";

static XS (menu_add)
{
	static const char function []= "menu_add";
	qpMenuAddParam param;
	int n;
	dXSARGS;
	if (items != 2)
	{
		require_argument (function, 2);
		XSRETURN_UNDEF;
	}
	param.position= SvIV (ST (0) );
	param.menuname= SvPV (ST (1), PL_na);
	n= qpCallback (gdata, qpMenuAdd, 0, & param);
	XSRETURN_IV (n);
}

static const char help_menu_add_option []=
"Pendiente de escribir la ayuda";

static XS (menu_add_option)
{
	static const char function []= "menu_add_option";
	qpMenuAddOptionParam param;
	int n;
	dXSARGS;
	if (items != 4)
	{
		require_argument (function, 4);
		XSRETURN_UNDEF;
	}
	param.menuname= SvPV (ST (0), PL_na);
	param.position= SvIV (ST (1) );
	param.optionname= SvPV (ST (2), PL_na);
	param.perlfunc= SvPV (ST (3), PL_na);
	n= qpCallback (gdata, qpMenuAddOption, 0, & param);
	XSRETURN_IV (n);
}

static const char help_file_load []=
"Pendiente de escribir la ayuda";

static XS (file_load)
{
	static const char function []= "file_load";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileLoad, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_file_new []=
"Pendiente de escribir la ayuda";

static XS (file_new)
{
	static const char function []= "file_new";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileNew, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_file_save []=
"Pendiente de escribir la ayuda";

static XS (file_save)
{
	static const char function []= "file_save";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileSave, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_file_save_as []=
"Pendiente de escribir la ayuda";

static XS (file_save_as)
{
	static const char function []= "file_save_as";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileSaveAs, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_file_next []=
"Pendiente de escribir la ayuda";

static XS (file_next)
{
	static const char function []= "file_next";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileNext, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_file_close []=
"Pendiente de escribir la ayuda";

static XS (file_close)
{
	static const char function []= "file_close";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileClose, 0, NULL);
	XSRETURN_IV (n);
}

static const char help_file_quit []=
"Pendiente de escribir la ayuda";

static XS (file_quit)
{
	static const char function []= "file_quit";
	int n;
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	n= qpCallback (gdata, qpFileQuit, 0, NULL);
	XSRETURN_IV (n);
}

void do_help_add_func (const char * func, const char * desc)
{
	qpHelpAddFuncParam param;
	param.func= func;
	param.desc= desc;
	qpCallback (gdata, qpHelpAddFunc, 0, & param);
}

static const char help_help_add_func []=
"Pendiente de escribir la ayuda";

static XS (help_add_func)
{
	static const char function []= "help_add_func";
	dXSARGS;
	if (items != 2)
	{
		require_argument (function, 2);
		XSRETURN_UNDEF;
	}
	do_help_add_func (SvPV (ST (0), PL_na), SvPV (ST (1), PL_na) );
	XSRETURN_YES;
}

static const char help_macro []=
"Pendiente de escribir la ayuda";

static XS (macro)
{
	static const char function []= "macro";
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	qpCallback (gdata, qpMacro, 0, NULL);
	XSRETURN_YES;
}

static const char help_end_init []=
"Pendiente de escribir la ayuda";

static XS (end_init)
{
	static const char function []= "macro";
	dXSARGS;
	if (items != 0)
	{
		no_argument (function);
		XSRETURN_UNDEF;
	}
	init_ended= 1;
	XSRETURN_YES;
}

static const char help_stupid_test []=
"nicamente para pruebas internas, no usar";

static XS (stupid_test)
{
	dXSARGS;
	/*fprintf (stderr, "GIMME_V is %i\n", GIMME_V);*/
	qpCallback (gdata, qpFailTest, 0, NULL);
	XSRETURN (1);
}

typedef XS ( (*perlfunction) );

struct table_t {
	const char * const name;
	perlfunction func;
	const char * helptext;
};

#define def(a) { #a, a, help_ ## a }

struct table_t table []= {
	def (mode),
	def (empty),
	def (modified),
	def (up),
	def (down),
	def (left),
	def (right),
	def (home),
	def (end),
	def (pageup),
	def (pagedown),
	def (goto_line),
	def (get_char),
	def (get_word),
	def (delete_char),
	def (backspace),
	def (delete_line),
	def (replace_line),
	def (insert_char),
	def (insert),
	def (get_beginsel),
	def (get_endsel),
	def (get_selection),
	def (get_clipboard),
	def (set_clipboard),
	def (line),
	def (num_lines),
	def (char_line),
	def (get_line),
	def (get_pos),
	def (get_col),
	def (get_fileext),
	def (get_filename),
	def (status),
	def (status_question),
	def (dialog_get_string),
	def (dialog_show_text),
	def (define_control),
	def (redraw),
	def (quiet),
	def (toggle_select),
	def (end_select),
	def (select_all),
	def (cut),
	def (copy),
	def (paste),
	def (get_key),
	def (autoindent),
	def (autocr),
	def (menu_add),
	def (menu_add_option),
	def (file_load),
	def (file_new),
	def (file_save),
	def (file_save_as),
	def (file_next),
	def (file_close),
	def (file_quit),
	def (help_add_func),
	def (macro),
	def (end_init),
	def (stupid_test),
};

static void loadqtrefuncs ()
{
	static char module []= "qperl.c";
	static char perlfuncname [128];
	char * func;
	struct table_t * pt;
	int i, l;
	for (i= 0, pt= table; i < DIM (table); ++i, ++pt)
	{
		strcpy (perlfuncname, qtre_prefix);
		strcat (perlfuncname + 6, pt->name);
		newXS (perlfuncname, pt->func, module);
		strcpy (perlfuncname, qtre_prefix_default);
		strcat (perlfuncname + 6, pt->name);
		newXS (perlfuncname, pt->func, module);
		do_help_add_func (pt->name, pt->helptext);
	}
}

static int isCurrentDir (const char * str)
{
	char cur [PATH_MAX];
	if (! getcwd (cur, sizeof (cur) ) )
		return 0;
	return strcmp (cur, str) == 0;
}

static int checkerror (const char * title)
{
	if (! SvTRUE (ERRSV) )
		return 0;
	else
	{
		/*errormessage (title, (char *) SvPV_nolen (ERRSV) );*/
		/* SvPV_nolen not available in all versions. */
		STRLEN len;
		char * strerr;
		strerr= SvPV (ERRSV, len);
		errormessage (title, strerr);
		/* Set to false the quick way.*/
		SvCUR_set (ERRSV, 0);
		return 1;
	}
}

#ifdef __GCC__
extern void xs_init (pTHXo);
#else
extern void xs_init ();
#endif

static PerlInterpreter *MyPerl;

int qperl_init (char * * env, void * data)
{
	static char * pl_argv [] = { "", "-e", "1;" };
	char * strHome;
	struct stat statbuf;

	gdata= data;

	MyPerl= perl_alloc ();
	perl_construct (MyPerl);
	perl_parse (MyPerl, xs_init, DIM (pl_argv), pl_argv, env);

	loadqtrefuncs ();

	/* Option to not load initialization files. */
	if (getenv ("QTRE_NO_INIT") != NULL)
		return 0;

	/*
		Load the initilization files.
	*/

	/* First, global. */
	if (stat (strGlobalPl, & statbuf) == 0)
	{
		perl_require_pv ( (char *) strGlobalPl);
		if (checkerror ("Error in /etc/qtre.pl") )
		{
			qperl_destroy ();
			return 1;
		}
	}
	// /etc/qtre.pl can forbid further initialisation.
	if (init_ended)
		return 1;

	/* Second, user. */
	strHome= getenv ("HOME");
	if (strHome)
	{
		char strPackage [PATH_MAX];
		int l;
		l= strlen (strHome);
		strcpy (strPackage, strHome);
		if (l > 0 && strPackage [l - 1] != '/')
			strcat (strPackage, "/");
		strcat (strPackage, strPl);
		if (stat (strPackage, & statbuf) == 0)
		{
			perl_require_pv (strPackage);
			if (checkerror ("Error in user .qtre.pl") )
			{
				qperl_destroy ();
				return 1;
			}
		}
	}
	// $HOME/.qtre.pl can also forbid further initialisation,
	if (init_ended)
		return 0;

	/* Last, current directory if is not home */
	if (! strHome || ! isCurrentDir (strHome) )
	{
		/* Last chance to avoid current directory personlization:
		environment variable. */
		if (getenv ("QTRE_NO_DIR") )
			return 0;

		if (stat (strPl, & statbuf) == 0)
		{
			perl_require_pv ( (char *) strPl);
			if (checkerror ("Error in local .qtre.pl") )
			{
				qperl_destroy ();
				return 1;
			}
		}
	}
	return 0;
}

void qperl_destroy (void)
{
	perl_destruct (MyPerl);
	perl_free (MyPerl);
}

static CV * get_cv_function (const char * function)
{
	char * f;
	int l;
	CV * cv;

	l= strlen (function) + sizeof (qtre_prefix);
	f= malloc (l);
	if (! f)
		return NULL;
	strcpy (f, qtre_prefix);
	strcat (f, function);
	cv= perl_get_cv (f, 0);
	if (! cv)
		cv= perl_get_cv (f + sizeof (qtre_prefix) - 1, 0);
	free (f);
	return cv;
}

qpfunctionresult qperl_function (const char * function, void * data)
{
	qpfunctionresult r;
	/*char * f;*/
	CV * cv;
	dSP;
	PUSHMARK (SP);

	gdata= data;
	/*
	r= strlen (function) + sizeof (qtre_prefix);
	f= malloc (r);
	if (f == NULL)
		return qpfunction_notfound;
	strcpy (f, qtre_prefix);
	strcat (f, function);

	// Search for function, first with qtre prefix, if not
	// found withou prefix.
	cv= perl_get_cv (f, 0);
	if (cv == NULL)
		cv= perl_get_cv (f + sizeof (qtre_prefix) - 1, 0);
	*/
	cv= get_cv_function (function);

	if (cv != NULL)
	{
		/*perl_call_pv (f,
			G_VOID | G_DISCARD | G_NOARGS | G_EVAL);
		*/
		perl_call_sv ( (SV *) cv,
			G_VOID | G_DISCARD | G_NOARGS | G_EVAL);
		if (checkerror ("Error in function") )
			r= qpfunction_failed;
		else
			r= qpfunction_ok;
	}
	else r= qpfunction_notfound;
	/*free (f);*/
	return r;
}

qpfunctionresult qperl_function_args (const char * function, void * data,
	int argc, const char * const * argv)
{
	qpfunctionresult r;
	CV * cv;
	int i;

	dSP;
	ENTER;
	SAVETMPS;
	PUSHMARK (SP);
	for (i= 0; i < argc; ++i)
	{
		/* Casting required in some versions to avoid warning */
		char * str= (char *) argv [i];
		XPUSHs (sv_2mortal (newSVpv (str, 0) ) );
	}
	PUTBACK;

	r= qpfunction_ok;
	gdata= data;
	cv= get_cv_function (function);
	if (! cv)
		r= qpfunction_notfound;
	else
	{
		perl_call_sv ( (SV *) cv,
			G_VOID | G_DISCARD | G_EVAL);
		if (checkerror ("Error in function") )
			r= qpfunction_failed;
	}

	SPAGAIN;
	PUTBACK;
	FREETMPS;
	LEAVE;
	return r;
}

int qperl_eval (const char * str, void * data)
{
	char * strerr;
	SV * svstr;
	dSP;
	PUSHMARK (SP);
	gdata= data;

	/* Need perl_eval_sv because perl_eval_pv does not report
	 syntax errors.
	 */
	svstr= newSVpv ( (char *) str, 0);
	perl_eval_sv (svstr,
		G_VOID | G_DISCARD | G_NOARGS | G_EVAL | G_KEEPERR);
	/*sv_2mortal (svstr); */
	sv_unref (svstr);

	return checkerror ("Error in eval");
}

/* Fin de qperl.c */
