// qtre.cpp
// Revision 12-mar-2004

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <stdexcept>
#include <memory>
#include <algorithm>
#include <vector>
//#include <map>

#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "termstream.h"
#include "text.h"
#include "qperl.h"
#include "menu.h"
#include "dialog.h"
#include "trace.h"
#include "exec.h"
#include "util.h"
#include "dirutil.h"
#include "status.h"
#include "history.h"
#include "select.h"
#include "selectfile.h"
#include "finder.h"

using util::dim_array;

#include "qtreassert.h"

using std::string;
using std::flush;
using std::auto_ptr;

extern const string strVersion;

bool fExit= false;

class TextWindow
{
private:
	size_t * pcount;
	Text * ptext;
public:
	TextWindow ();
	TextWindow (const TextWindow & tw);
	~TextWindow ();
	void operator = (const TextWindow & tw);
	Text * text () { return ptext; }
	size_t lin;
	Col col;
	size_t inilin;
	Col inicol;
	Selection sel;
};

TextWindow::TextWindow () :
	pcount (new size_t),
	ptext (new Text),
	lin (0),
	col (0),
	inilin (0),
	inicol (0)
{
	TraceFunc tr ("TextWindow::TextWindow");

	* pcount= 1;
}

TextWindow::TextWindow (const TextWindow & tw) :
	pcount (tw.pcount),
	ptext (tw.ptext),
	lin (tw.lin),
	col (tw.col),
	inilin (tw.inilin),
	inicol (tw.inicol),
	sel (tw.sel)
{
	TraceFunc tr ("TextWindow::TextWindow-copy");

	++ * pcount;
}

TextWindow::~TextWindow ()
{
	TraceFunc tr ("TextWindow::~TextWindow");

	if (-- * pcount == 0)
		delete ptext;
}

void TextWindow::operator = (const TextWindow & tw)
{
	TraceFunc tr ("TextWindow::operator =");

	if (this != & tw)
	{
		if (-- * pcount == 0)
			delete ptext;
		pcount= tw.pcount;
		++ * pcount;
		ptext= tw.ptext;
		lin= tw.lin;
		col= tw.col;
		inilin= tw.inilin;
		inicol= tw.inicol;
		sel= tw.sel;
	}
}

class Screen;

class ScreenAction : public Action {
public:
	ScreenAction (Screen & screen) :
		screen (screen)
	{ }
protected:
	Screen & screen;
};

class ScreenActionFunction : public ScreenAction {
public:
	ScreenActionFunction (Screen & screen, void (Screen::* function) () ) :
		ScreenAction (screen),
		function (function)
	{ }
	void doit ()
	{
		(screen.*function) ();
	}
private:
	void (Screen::* function) ();
};

class ScreenActionPerl : public ScreenAction {
public:
	ScreenActionPerl (Screen & screen, const std::string & function) :
		ScreenAction (screen),
		function (function)
	{ }
	void doit ();
private:
	const std::string function;
};

class Screen : public MapKeyAction, private Exec
{
	struct SearchOptions {
		bool nocase;
		bool backwards;
		bool wholeword;
		bool regex;
		SearchOptions () :
			nocase (false),
			backwards (false),
			wholeword (false),
			regex (false)
		{ }
	};
	//friend int qpCallback (void *, qpFuncCode, int, void *);
public:
	int callback (qpFuncCode f, int n, void * data);
	void showerror (const char * str);
	typedef std::vector <TextWindow> ContainerWindow;
private:
	ContainerWindow win;
	size_t actualwin;
	Text * ptext;
	auto_ptr <Text> pclipboard;
	StatusLine & status;
	Menu & menu;
	otermstream & ot;
	otermstream & mainwindow;
	otermstream & menuwindow;
	bool autoindent;
	bool autocr;
	size_t maxlin;
	Col maxcol;
	bool redraw_all;
	bool fSelecting;
	bool exiting;
	bool quietmode;

	string strFind;
	string strReplace;
	Finder * pfinder;
	History historyfind;
	History historyreplace;

	string strEval;
	History historyfilename;
	SearchOptions lastsearchoptions;
	string strPrompt;
	History historyprompt;
	string strCompileOptions;
	History historycompileoptions;
	string strMan;
	History historyman;
	typedef std::map <std::string, std::string> perlfunchelp_t;
	perlfunchelp_t perlfunchelp;

	void ExecedSendChar (char c)
	{
		sendchar (c);
	}
public:
	Screen (StatusLine & st, Menu & mn,
		otermstream & ot, otermstream & mainwindow,
		otermstream & menuwindow);
	~Screen ();
	bool setmode (int n);
	int getkey ()
	{ return ot.getkey (); }
	void setstatus (const std::string & str)
	{ status.settext (str); }
	int statusgetkey (const std::string & str);
	void redraw ();
	void windownew ();
	void nextwindow ();
	size_t & lin () { return win [actualwin].lin; }
	Col & col () { return win [actualwin].col; }
	size_t getpos () { return (* ptext) [lin () ].col2pos (col () ); }
	size_t & inilin () { return win [actualwin].inilin; }
	Col & inicol () { return win [actualwin].inicol; }
	void beginselect ();
	void endselect ();
	void selectall ();
	void copyselection (Text & clipboard);
	void copyselection ();
	void copyselection (char * str);
	bool copyselection (char * & str, size_t & len);
	void copyclipboard (char * str);
	bool copyclipboard (char * & str, size_t & len);
	int lengthselection ();
	int lengthclipboard ();
	void deleteselection ();
	void cut ();
	void copy ();
	void paste ();
	void gotoline (size_t l, Col gotocol= Col () );
	char getcurchar ();
	void getline (int nline, char * str);
	bool getline (int nline, char * & str, size_t & len);
	void getfileext (int n, char * str);
	void getfilename (int n, char * str);
	void instext (const char * str);
	void do_instext (const char * str, size_t len);
	void instext (const char * str, size_t len);
	void putline (size_t l);
	inline void putline (const PosText & pt);
	inline void putlineabs (size_t i);
	inline void putactualline ();
	void putall ();
	void scrollup ();
	void scrolldown ();
	void adjustmaxcol ();
	void movesel (const PosText & opt, size_t lin, size_t pos);
	void home ();
	void end ();
	void inicio ();
	void fin ();
	bool up ();
	void move_up ();
	void pageup ();
	bool down ();
	void move_down ();
	void pagedown ();
	bool back ();
	void move_left ();
	bool forward ();
	void move_right ();
	void deletechar ();
	void backdelete ();
	void inschar (unsigned char c);
	void sendchar (unsigned char c);
	void adjustinicol ();
	void adjustinilin ();
	bool do_save ();
	bool do_save_as ();
	bool do_save_all ();
	void do_execute (const std::string & str);
	bool question_save ();

	void optionLoad ();
	void optionSave ();
	void optionSaveAs ();
	void optionQuit ();
	void optionQuitOrClose ();
	void optionCloseWindow ();
	void optionFileSelect ();
	void optionFileOptions ();

	void optionEditOptions ();

	bool askSearchOptions (SearchOptions & options);
	void optionSelect ();
	void buildfinder (const SearchOptions & options,
		const std::string & strFind);
	void optionFind ();
	bool searchtext (const std::string & searched, SearchOptions options);
	void optionFindAgain ();
	void optionReplace ();

	void optionGoLine ();

	void optionMan ();
	void optionManOf ();
	void openman (const std::string & str);
	void optionShell ();
	void optionExecute ();
	void optionExecuteCurrent ();
	void optionEvaluate ();
	void optionEvalSelection ();
	void optionEvalClipboard ();
	void optionEvalCurrent ();
	void optionDefineControl ();

	void optionAutoIndent ();

	void optionCompile ();
	void optionMake ();
	void optionCompileOptions ();

	void optionPrompt ();

	void optionHelp ();
	void optionHelpPerl ();
	void optionAbout ();

	void perlfunc (const std::string & func);
	void macro ();

	bool fastkey (int c);
	void definecontrolkey (char k, const std::string & function);
	void menuselect ();
	void sendkey (int c);
	void deleteline ();
	void replaceline (char * str, int line);
	void interact ();
	void loadtext (const string & strFileName);
	void changewindow (size_t newwin);
private:
	void inittext ();
	ContainerWindow initwin ();
	bool closewindow ();
	void init_actions ();

	#if 0
	class MapKeyAction {
	public:
		~MapKeyAction ()
		{
			for (mapka::iterator it= m.begin (), end= m.end ();
				it != end; ++it)
			{
				delete it->second;
			}
		}
		void add (int key, Action * action)
		{
			mapka::iterator it= m.find (key);
			if (it != m.end () )
				delete it->second;
			m [key]= action;
		}
		Action * get (int key)
		{
			mapka::iterator it= m.find (key);
			if (it != m.end () )
				return it->second;
			else
				return 0;
		}
	private:
		typedef std::map <int, Action *> mapka;
		mapka m;
	};
	MapKeyAction mapkeyaction;
public:
	void addKeyAction (int key, Action * action)
	{
		mapkeyaction.add (key, action);
	}
	#endif
};

void ScreenActionPerl::doit ()
{
	screen.perlfunc (function);
}

Screen::Screen (StatusLine & st, Menu & mn,
		otermstream & ot, otermstream & mainwindow,
		otermstream & menuwindow) :
	win (initwin () ),
	actualwin (0),
	ptext (win [actualwin].text () ),
	pclipboard (new Text),
	status (st),
	menu (mn),
	ot (ot),
	mainwindow (mainwindow),
	menuwindow (menuwindow),
	autoindent (true),
	autocr (false),
	maxlin (ot.getHeight () - 1),
	maxcol (ot.getWidth () -1),
	exiting (false),
	quietmode (false),
	pfinder (0)
{
	TraceFunc tr ("Screen::Screen");

	inittext ();
	setmode (0);
	init_actions ();
}

Screen::~Screen ()
{
	TraceFunc tr ("Screen::~Screen");
}

bool Screen::setmode (int n)
{
	switch (n)
	{
	case 0:
		ot.setcolor (2, Color::White, Color::Black);
		ot.color (2);
		ot.background (2);
		return true;
	case 1:
		ot.setcolor (2, Color::Black, Color::White);
		ot.color (2);
		ot.background (2);
		return true;
	default:
		return false;
	}
}

int Screen::statusgetkey (const std::string & str)
{
	return status.getkey (str);
}

void Screen::redraw ()
{
	ot << clrscr;
	putall ();
	status.refresh ();
}

Screen::ContainerWindow Screen::initwin ()
{
	ContainerWindow aux;
	aux.push_back (TextWindow () );
	return aux;
}

void Screen::changewindow (size_t newwin)
{
	TraceFunc tr ("Screen::changewindow");

	endselect ();
	actualwin= newwin;
	ASSERT (actualwin < win.size () );
	ptext= win [actualwin].text ();
	if (quietmode)
		redraw_all= true;
	else
		putall ();
}

void Screen::windownew ()
{
	win.push_back (TextWindow () );
	changewindow (win.size () - 1);
}

void Screen::nextwindow ()
{
	changewindow ( (actualwin + 1) % win.size () );
}

bool Screen::closewindow ()
{
	TraceFunc tr ("Screen::closewindow");

	if (! question_save () )
		return false;
	win.erase (win.begin () + actualwin);
	if (win.size () == 0 && ! exiting)
		win.push_back (TextWindow () );
	size_t newwin= actualwin;
	if (newwin >= win.size () )
		newwin= 0;
	if (win.size () > 0)
		changewindow (newwin);
	return true;
}

int Screen::callback (qpFuncCode f, int n, void * data)
{
	TraceFunc tr ("Screen::callback");

	switch (f)
	{
	case qpMode:
		return setmode (n);
	case qpIsEmpty:
		return ptext->empty ();
	case qpGetModified:
		return ptext->modified ();
	case qpSetModified:
		{
			int r= ptext->modified ();
			ptext->modified (n);
			return r;
		}
	case qpUp:
		return up ();
	case qpDown:
		return down ();
	case qpLeft:
		return back ();
	case qpRight:
		return forward ();
	case qpHome:
		home ();
		break;
	case qpEnd:
		end ();
		break;
	case qpPageUp:
		pageup ();
		break;
	case qpPageDown:
		pagedown ();
		break;
	case qpGotoLine:
		if (static_cast <unsigned> (n) >= ptext->lines () )
			return 0;
		gotoline (n);
		break;
	case qpGetChar:
		return getcurchar ();
		break;
	case qpGetWord:
		{
			qpStringParam * pparam=
				static_cast <qpStringParam *> (data);
			std::string str
				(ptext->wordat (lin (), col () ) );
			pparam->str= new (std::nothrow) char [str.size () + 1];
			if (! pparam->str)
				return 0;
			pparam->len= str.size ();
			str.copy (pparam->str, str.size () );
			pparam->str [str.size () ]= '\0';
		}
		break;
	case qpDeleteChar:
		deletechar ();
		break;
	case qpBackspace:
		backdelete ();
		break;
	case qpDeleteLine:
		deleteline ();
		break;
	case qpReplaceLine:
		replaceline ( (char *) data, n);
		break;
	case qpInsertChar:
		sendchar (n);
		break;
	case qpInsert:
		//instext ( (char *) data);
		{
			qpStringParam * pparam=
				static_cast <qpStringParam *> (data);
			instext (pparam->str, pparam->len);
		}
		break;
	case qpLengthSelection:
		return lengthselection ();
	case qpGetSelection:
		//copyselection ( (char *) data);
		//break;
		{
			qpStringParam * pparam=
				static_cast <qpStringParam *> (data);
			return copyselection (pparam->str, pparam->len);
		}
	case qpSelectionLimits:
		{
			Selection & sel= win [actualwin].sel;
			if (sel.begin () == sel.end () )
				return 0;
			qpSelectionLimitsParam * pparam=
				static_cast <qpSelectionLimitsParam *> (data);
			pparam->beginline= sel.begin ().lin;
			pparam->beginpos= sel.begin ().pos;
			pparam->endline= sel.end ().lin;
			pparam->endpos= sel.end ().pos;
			return 1;
		}
	case qpLengthClipboard:
		return lengthclipboard ();
	case qpGetClipboard:
		//copyclipboard ( (char *) data);
		//break;
		{
			qpStringParam * pparam=
				static_cast <qpStringParam *> (data);
			return copyclipboard (pparam->str, pparam->len);
		}
	case qpSetClipboard:
		tr.message ( (char *) data);
		pclipboard->set ( (char *) data);
		break;
	case qpLine:
		return lin ();
	case qpNumLines:
		return ptext->lines ();
	case qpCharLine:
		return (* (ptext) ) [n].size ();
	case qpGetLine:
		//getline (n, (char *) data);
		//break;
		{
			qpStringParam * pparam=
				static_cast <qpStringParam *> (data);
			return getline (n, pparam->str, pparam->len);
		}
	case qpGetPos:
		return getpos ();
	case qpGetCol:
		return col ().getn ();
	case qpGetFileExt:
		getfileext (n, (char *) data);
		break;
	case qpGetFileName:
		getfilename (n, (char *) data);
		break;
	case qpStatus:
		status.settext ( (char *) data);
		break;
	case qpStatusQuestion:
		{
			qpStatusQuestionParam * pparam=
				static_cast <qpStatusQuestionParam *> (data);
			return status.question
				(pparam->question, pparam->values);
		}
	case qpDialogGetString:
		{
			qpDialogGetStringParam * pparam=
				static_cast <qpDialogGetStringParam *> (data);
			string str (pparam->str);
			static History history;
			int k;
			bool r= dialogGetString (ot,
				k, std::set <int> (),
				pparam->title, pparam->prompt,
				str, history,
				pparam->width, pparam->maxlength);
			putall ();
			if (r)
			{
				std::string::size_type l= str.size ();
				ASSERT (l <= pparam->maxlength);
				str.copy (pparam->str, l);
				pparam->str [l]= '\0';
			}
			return r;
		}
	case qpDialogShowText:
		{
			qpDialogShowTextParam * pparam=
				static_cast <qpDialogShowTextParam *> (data);
			dialogShowText (ot,
				pparam->title, pparam->str,
				pparam->height, pparam->width);
			putall ();
		}
		break;
	case qpDefineControl:
		{
			qpDefineControlParam * pparam=
				static_cast <qpDefineControlParam *> (data);
			std::string function (pparam->function);
			definecontrolkey (pparam->key, function);
		}
		break;
	case qpRedraw:
		redraw ();
		break;
	case qpQuiet:
		quietmode= n;
		break;
	case qpSelect:
		optionSelect ();
		break;
	case qpEndSelect:
		endselect ();
		break;
	case qpSelectAll:
		selectall ();
		break;
	case qpCut:
		cut ();
		break;
	case qpCopy:
		copy ();
		break;
	case qpPaste:
		paste ();
		break;
	case qpGetKey:
		return getkey ();
	case qpSetAutoindent:
		{
			int r= autoindent;
			autoindent= n != 0;
			return r;
		}
	case qpGetAutoindent:
		return autoindent;
	case qpSetAutocr:
		{
			int r= autocr;
			autocr= n != 0;
			return r;
		}
	case qpGetAutocr:
		return autocr;
	case qpMenuAdd:
		{
			qpMenuAddParam * pparam=
				static_cast <qpMenuAddParam *> (data);

			std::ostringstream oss;
			oss << pparam->position << ' ' << pparam->menuname;
			status.settext (oss.str () );

			Menu * pmenu= new Menu ();
			Option * pop= new Option (pparam->menuname,
				pmenu, MenuOwner);
			menu.insertOption (pop, MenuOwner,
				pparam->position);
			menu.show (menuwindow);
		}
		break;
	case qpMenuAddOption:
		{
			qpMenuAddOptionParam * pparam=
				static_cast <qpMenuAddOptionParam *> (data);
			std::ostringstream oss;

			oss << pparam->menuname << ' ' << pparam->position <<
				' ' << pparam->optionname << ' ' <<
				pparam->perlfunc;
			status.settext (oss.str () );

			Menu * pmenu= menu.getSubMenu (pparam->menuname);
			if (pmenu == NULL)
				return 0;
			ScreenActionPerl * paction= new ScreenActionPerl
				(* this, pparam->perlfunc);
			Option * pop= new Option
				(pparam->optionname, paction, MenuOwner);
			pmenu->insertOption (pop, MenuOwner, pparam->position);
		}
		break;
	case qpFileLoad:
		optionLoad ();
		break;
	case qpFileNew:
		windownew ();
		break;
	case qpFileSave:
		optionSave ();
		break;
	case qpFileSaveAs:
		optionSaveAs ();
		break;
	case qpFileNext:
		nextwindow ();
		break;
	case qpFileClose:
		optionCloseWindow ();
		break;
	case qpFileQuit:
		optionQuit ();
		break;
	case qpHelpAddFunc:
		{
			qpHelpAddFuncParam * pparam=
				static_cast <qpHelpAddFuncParam *> (data);
			perlfunchelp [pparam->func]= pparam->desc;
		}
		break;
	case qpMacro:
		macro ();
		break;
	case qpFreeString:
		{
			qpStringParam * pparam=
				static_cast <qpStringParam *> (data);
			delete [] pparam->str;
		}
		break;
	case qpFailTest:
		throw std::runtime_error ("This is a fail test");
	}
	return 1;
}

void Screen::showerror (const char * str)
{
	// Try by all means to show the error without
	// propagating exceptions.
	try
	{
		dialogShowText (ot, "ERROR", str,
			std::min (ot.lines () / 2, 7), ot.columns () / 2);
		putall ();
	}
	catch (...)
	{
		try
		{
			status.settext (std::string ("ERROR: ") + str);
		}
		catch (...)
		{
			std::cerr << "ERROR: " << str << std::endl;
		}
	}
}

int qpCallback (void *thisclass, qpFuncCode f, int n, void * data)
{
	TraceFunc tr ("qpCallback");

	// Ensure no exception is propagated back to qperl.

	Screen *p= static_cast <Screen *> (thisclass);
	try
	{
		return p->callback (f, n, data);
	}
	catch (std::exception & e)
	{
		p->showerror (e.what () );
	}
	catch (...)
	{
		p->showerror ("Unexpected exception");
	}
	return -1;
}

void Screen::getline (int nline, char * str)
{
	const TextLine & s= (* ptext) [nline];
	size_t l= s.size ();
	size_t i;
	for (i= 0; i < l; ++i)
		str [i]= s [i];
	str [i]= '\0';
}

bool Screen::getline (int nline, char * & str, size_t & len)
{
	const TextLine & s= (* ptext) [nline];
	len= s.size ();
	str= new (std::nothrow) char [len + 1];
	if (str == NULL)
		return false;
	s.str ().copy (str, len);
	str [len]= '\0';
	return true;
}

void Screen::getfileext (int n, char * str)
{
	TraceFunc tr ("Screen::getfileext");

	std::string ext (dirutil::extension (ptext->name () ) );
	std::string::size_type l;
	if ( (l= ext.size () ) < std::string::size_type (n) )
		ext.copy (str, l);
	else
		l= 0;
	str [l]= 0;
	tr.message (str);
}

void Screen::getfilename (int n, char * str)
{
	TraceFunc tr ("Screen::getfilename");

	std::string file (dirutil::base_name (ptext->name () ) );
	std::string::size_type l;
	if ( (l= file.size () ) < std::string::size_type (n) )
		file.copy (str, l);
	else
		l= 0;
	str [l]= 0;
	tr.message (str);
}

void Screen::instext (const char * str)
{
	util::bool_saver saveautoindent (autoindent, false);
	util::bool_saver saveautocr (autocr, false);
	for (; * str != '\0'; ++str)
		sendchar (* str);
}

namespace {

const char * findnewline (const char * str, size_t len)
{
	for ( ; len > 0; --len, ++str)
		if (* str == '\n')
			return str;
	return 0;
}

} // namespace

void Screen::do_instext (const char * str, size_t len)
{
	util::bool_saver saveautoindent (autoindent, false);
	util::bool_saver saveautocr (autocr, false);
	size_t initial= lin ();
	size_t pos= getpos ();
	const char * straux= findnewline (str, len);
	if (! straux)
	{
		const std::string & s= (* ptext) [initial].str ();
		ptext->setline (initial, s.substr (0, pos) +
			std::string (str, len) + s.substr (pos) );
		pos+= len;
		col ()= (* ptext) [initial].pos2col (pos);
		adjustinicol ();
		if (! redraw_all)
			putlineabs (initial);
	}
	else
	{
		//for (; len > 0; ++str, --len)
		//	sendchar (* str);

		ptext->insertline (initial, col (), false);
		size_t l= straux - str;
		ptext->append (initial, std::string (str, l) );
		size_t current= initial + 1;
		str= straux + 1;
		len-= l + 1;

		while ( (straux= findnewline (str, len) ) != 0)
		{
			l= straux - str;
			ptext->insertline (current, std::string (str, l) );
			str= straux + 1;
			len-= l + 1;
			++current;
		}

		const std::string & s= (* ptext) [current].str ();
		ptext->setline (current, std::string (str, len) + s);

		lin ()= current;
		col ()= (* ptext) [current].pos2col (len);
		adjustinicol ();
		adjustinilin ();
		redraw_all= true;
	}
}

void Screen::instext (const char * str, size_t len)
{
	try
	{
		do_instext (str, len);
	}
	catch (std::exception &e)
	{
		status.settext (std::string ("Error: ") + e.what () );
	}
	catch (...)
	{
		status.settext ("Unexpected error");
	}
}

void Screen::inittext ()
{
	inilin ()= 0;
	inicol ()= Col ();
	lin ()= 0;
	col ()= Col ();
	fSelecting= false;
}

void Screen::beginselect ()
{
	status.settext ("Seleccionando");
	Selection & sel= win [actualwin].sel;
	bool fSelPrev= sel.begin () != sel.end ();

	//const TextLine & s= (* ptext) [lin () ];
	//size_t pos= s.col2pos (col () );
	//sel.init (lin (), pos);
	sel.init (lin (), getpos () );

	fSelecting= true;
	if (fSelPrev)
		redraw_all= true;
}

void Screen::endselect ()
{
	status.clear ();
	fSelecting= false;
}

void Screen::selectall ()
{
	TraceFunc tr ("Screen::selectall");

	Selection & sel= win [actualwin].sel;
	tr.message ("Moving begin");
	sel.movebegin (0, 0);
	tr.message ("Evaulating last line");
	size_t n= ptext->lines ();
	if (n == 0)
		sel.moveend (0, 0);
	else
	{
		tr.message ("Evaluating last col");
		size_t pos= (* ptext) [n - 1].size ();
		tr.message ("Moving end");
		sel.moveend (n - 1, pos);
	}
	fSelecting= false;
	redraw_all= true;
}

void Screen::copyselection (Text & clipboard)
{
	endselect ();
	Selection & sel= win [actualwin].sel;
	if (sel.begin () == sel.end () )
		return;
	clipboard.clear ();
	size_t linfirst= sel.begin ().lin;
	size_t linlast= sel.end ().lin;
	size_t posfirst= sel.begin ().pos;
	size_t poslast= sel.end ().pos;
	string s;
	if (linfirst == linlast)
	{
		s= (* ptext) [linfirst].substr (posfirst, poslast - posfirst);
		clipboard.setline (0, s);
	}
	else
	{
		s= (* ptext) [linfirst].substr (posfirst);
		clipboard.setline (0, s);
		for (size_t i= linfirst + 1; i < linlast; ++i)
			clipboard.push_back ( (* ptext) [i].str () );
		s= (* ptext) [linlast].substr (0, poslast);
		clipboard.push_back (s);
	}
}

void Screen::copyselection ()
{
	copyselection (* pclipboard);
}

void Screen::copyselection (char * str)
{
	TraceFunc tr ("Screen::copyselection (char *str)");

	Text clipboard;
	copyselection (clipboard);
	clipboard.c_str (str);
}

bool Screen::copyselection (char * & str, size_t & len)
{
	Text clipboard;
	copyselection (clipboard);
	return clipboard.copy (str, len);
}

void Screen::copyclipboard (char * str)
{
	TraceFunc tr ("Screen::copyclipboard");

	pclipboard->c_str (str);
}

bool Screen::copyclipboard (char * & str, size_t & len)
{
	return pclipboard->copy (str, len);
}

int Screen::lengthselection ()
{
	Text clipboard;
	copyselection (clipboard);
	return clipboard.size ();
}

int Screen::lengthclipboard ()
{
	return pclipboard->size ();
}

void Screen::deleteselection ()
{
	endselect ();
	Selection & sel= win [actualwin].sel;
	if (sel.begin () == sel.end () )
		return;
	size_t linbeg= sel.begin ().lin;
	size_t linend= sel.end ().lin;
	Text & text= * ptext;
	if (linbeg == linend)
		text.setline (linbeg,
			text [linbeg].substr (0, sel.begin ().pos) +
			text [linbeg].substr (sel.end ().pos) );
	else
	{
		text.setline (linbeg,
			text [linbeg].substr (0, sel.begin ().pos) );
		text.setline (linend,
			text [linend].substr (sel.end ().pos) );
		for (size_t lin= linbeg + 1; lin < linend; ++lin)
			text.deleteline (linbeg + 1);
		text.joinline (linbeg);
	}
	lin ()= linbeg;
	col ()= text [lin () ].pos2col (sel.begin ().pos);
	sel.init (0, 0);
	adjustinilin ();
	redraw_all= true;
}

void Screen::cut ()
{
	endselect ();
	Selection & sel= win [actualwin].sel;
	if (sel.begin () == sel.end () )
		return;
	status.settext ("Cortando");
	copyselection ();
	deleteselection ();
}

void Screen::copy ()
{
	endselect ();
	Selection & sel= win [actualwin].sel;
	if (sel.begin () == sel.end () )
		return;
	status.settext ("Copiando");
	copyselection ();	
}

void Screen::paste ()
{
	endselect ();
	status.settext ("Pegando");
	util::bool_saver saveautoindent (autoindent, false);
	util::bool_saver saveautocr (autocr, false);
	util::bool_saver savequietmode (quietmode, true);

	size_t l= pclipboard->lines ();
	size_t current= lin ();
	size_t pos= getpos ();
	if (l == 1)
	{
		const std::string & s= (* ptext) [current].str ();
		const std::string & str= (* pclipboard) [0].str ();
		ptext->setline (current, s.substr (0, pos) + str +
			s.substr (pos) );
		pos+= str.size ();
		col ()= (* ptext) [current].pos2col (pos);
		adjustinicol ();
		if (! redraw_all)
			putlineabs (current);
	}
	else
	{
		#if 0
		for (size_t i= 0; i < l; ++i)
		{
			const TextLine & s= pclipboard->operator [] (i);
			for (size_t j= 0; j < s.size (); ++j)
				sendchar (s [j]);
			if (i != l - 1)
				sendchar ('\n');			
		}
		#endif

		ptext->insertline (current, col (), false);
		ptext->append (current, (* pclipboard) [0].str () );
		++current;

		for (size_t i= 1; i < l - 1; ++i, ++current)
		{
			ptext->insertline
				(current, (* pclipboard) [i].str () );
		}

		const std::string & s= (* ptext) [current].str ();
		const std::string & lastclip= (* pclipboard) [l - 1].str ();
		ptext->setline (current, lastclip + s);

		lin ()= current;
		col ()= (* ptext) [current].pos2col (lastclip.size () );
		adjustinicol ();
		adjustinilin ();
		redraw_all= true;
	}
	status.clear ();
}

void Screen::gotoline (size_t l, Col gotocol)
{
	TraceFunc tr ("Screen::gotoline");

	//if (l >= ptext->lines () )
	//	return;
	ASSERT (l < ptext->lines () );
	PosText pt;
	if (fSelecting)
		pt.set (lin (), getpos () );
	size_t iniant= inilin ();
	if (l < inilin () )
	{
		lin ()= l;
		inilin ()= l;
	}
	else
		if (l < inilin () + maxlin)
			lin ()= l;
		else
		{
			inilin ()= l - maxlin;
			lin ()= l;
		}
	col ()= gotocol;
	if (fSelecting)
		movesel (pt, lin (), getpos () );
	if (inilin () != iniant || fSelecting)
		redraw_all= true;
	adjustinicol ();
}

char Screen::getcurchar ()
{
	const TextLine & s= (* ptext) [lin () ];
	size_t pos= s.col2pos (col () );
	if (pos >= s.size () )
	{
		if (lin () >= ptext->lines () - 1) return '\0';
		else return '\n';
	}
	return s [pos];
}

void Screen::putline (size_t l)
{
	ASSERT (l <= maxlin);
	if (redraw_all)
		return;
	size_t lin= l + inilin ();
	ot << at (l, 0);
	if (lin >= ptext->lines () )
	{
		ot << clreol;
	}
	else
	{
		size_t posBegSel= 0, posEndSel= 0;
		// Inicializadas para evitar un warning.
		Selection & sel= win [actualwin].sel;
		bool fSelected= lin >= sel.begin ().lin &&
			lin <= sel.end ().lin && sel.begin () != sel.end ();
		bool fSelectedBeg= lin == sel.begin ().lin;
		if (fSelectedBeg)
			posBegSel= sel.begin ().pos;
		bool fSelectedEnd= lin == sel.end ().lin;
		if (fSelectedEnd)
			posEndSel= sel.end ().pos;
		if (fSelected && ! fSelectedBeg)
			ot << inverse (true);
		const TextLine & s= (* ptext) [lin];
		size_t lim= s.size ();
		size_t pos= s.col2pos (inicol () );
		Col ini;
		if (pos <= lim)
		{
			ini= s.pos2col (pos);
			ASSERT (ini == inicol () || ini == inicol () + 1);
		}
		unsigned char c;
		if (ini.odd () && pos <= lim)
		{
			ASSERT (ini == inicol () + 1);
			ASSERT (pos > 0);
			ASSERT (Text::escapable (s [pos - 1] ) );
			ot << Text::escaped (s [pos - 1] );
		}
		Col acol;
		for (; acol <= maxcol && pos < lim; ++pos)
		{
			if (fSelectedBeg && pos >= posBegSel)
			{
				ot << inverse (true);
				fSelectedBeg= false;
			}
			if (fSelectedEnd && pos >= posEndSel)
			{
				ot << inverse (false);
				fSelectedEnd= false;
			}
			c= s [pos];
			if (c == '\t')
			{
				acol.nexttab ();
				ot << '\t';
			}
			else
				if (unsigned char escchar=
					Text::escapable (c) )
				{
					acol+= 2;
					//ot << '^';
					ot << escchar;
					if (acol <= maxcol + 1)
						ot << Text::escaped (c);
				}
				else
				{
					++acol;
					ot << c;
				}
		}
		if (acol <= maxcol)
			ot << clreol;
		ot << flush << inverse (false);
	}
}

inline void Screen::putlineabs (size_t line)
{
	size_t ini= inilin ();
	ASSERT (line >= ini);
	if (line < ini)
		return;
	size_t l= line - ini;
	ASSERT (l <= maxlin);
	if (l > maxlin)
		return;
	putline (l);
}

inline void Screen::putline (const PosText & pt)
{
	putlineabs (pt.lin);
}

inline void Screen::putactualline ()
{
	putlineabs (lin () );
}

void Screen::putall ()
{
	TraceFunc tr ("Screen::putall");

	redraw_all= false;
	if (win.size () == 0)
		return;
	for (size_t i= 0; i <= maxlin; ++i)
		putline (i);
}

void Screen::scrollup ()
{
	--inilin ();
	if (quietmode)
		redraw_all= true;
	else
	{
		ot << at (0, 0) << insline;
		putline (0);
	}
}

void Screen::scrolldown ()
{
	++inilin ();
	if (quietmode)
		redraw_all= true;
	else
	{
		ot << at (0, 0) << delline;
		putline (maxlin);
	}
}

void Screen::adjustmaxcol ()
{
	const TextLine & s= (* ptext) [lin () ];
	Col m= s.colmax ();
	if (col () > m)
		col ()= m;
	else
		col()= s.pos2col (s.col2pos (col () ) );
}

void Screen::movesel (const PosText & opt, size_t lin, size_t pos)
{
	PosText npt (lin, pos);
	Selection & sel= win [actualwin].sel;
	if (opt == sel.end () )
		sel.moveend (npt);
	else
	{
		ASSERT (opt == sel.begin () );
		sel.movebegin (npt);
	}
}

void Screen::home ()
{
	if (col () == Col () )
		return;
	PosText opt;
	if (fSelecting)
		//opt.set (lin (), (* ptext) [lin () ].col2pos (col () ) );
		opt.set (lin (), getpos () );
	col ()= Col ();
	adjustinicol ();
	if (fSelecting)
	{
		movesel (opt, lin (), 0);
		if (! redraw_all)
			putactualline ();
	}
}

void Screen::end ()
{
	const TextLine & s= (* ptext) [lin () ];
	size_t npos= s.size ();
	Col acol (col () );
	col ()= s.pos2col (npos);
	if (col () == acol)
		return;
	adjustinicol ();
	if (fSelecting)
	{
		movesel (PosText (lin (), s.col2pos (acol) ), lin (), npos);
		if (! redraw_all)
			putactualline ();
	}
}

bool Screen::up ()
{
	if (lin () == 0)
		return false;
	PosText opt;
	Text & text= * ptext;
	if (fSelecting)
		opt.set (lin (), text [lin () ].col2pos (col () ) );
	--lin ();
	adjustmaxcol ();
	if (fSelecting)
	{
		size_t npos= text [lin () ].col2pos (col () );
		movesel (opt, lin (), npos);
		putline (opt);
	}
	if (lin () < inilin () )
		scrollup ();
	else
		if (fSelecting)
			putactualline ();
	return true;
}

void Screen::move_up ()
{
	up ();
}

void Screen::pageup ()
{
	size_t oini= inilin ();
	PosText opt;
	Text & text= * ptext;
	if (fSelecting)
		opt.set (lin (), text [lin () ].col2pos (col () ) );
	if (inilin () > 0)
	{
		if (inilin () > maxlin) inilin ()-= maxlin;
		else inilin ()= 0;
		if (lin () > maxlin) lin()-= maxlin;
		else lin ()= 0;
	}
	else
	{
		if (lin () == 0) return;
		lin ()= 0;
	}
	adjustmaxcol ();
	if (fSelecting)
		movesel (opt, lin (), text [lin () ].col2pos (col () ) );
	if (oini != inilin () || fSelecting)
		//putall ();
		redraw_all= true;
}

bool Screen::down ()
{
	Text & text= * ptext;
	if (lin () + 1 >= text.lines () ) return false;
	PosText opt;
	if (fSelecting)
		opt.set (lin (), text [lin () ].col2pos (col () ) );
	++lin ();
	adjustmaxcol ();
	if (fSelecting)
	{
		size_t npos= text [lin () ].col2pos (col () );
		movesel (opt, lin (), npos);
		putline (opt);
	}
	if (lin () > inilin () + maxlin)
	{
		scrolldown ();
		ASSERT (lin () == inilin () + maxlin);
	}
	else
		if (fSelecting)
			putactualline ();
	return true;
}

void Screen::move_down ()
{
	down ();
}

void Screen::pagedown ()
{
	Text & text= * ptext;
	if (inilin () >= text.lines () - 1)
	{
		ASSERT (inilin () == text.lines () - 1);
		return;
	}
	PosText opt;
	if (fSelecting)
		opt.set (lin (), text [lin () ].col2pos (col () ) );
	if (inilin () + maxlin < text.lines () )
	{
		inilin ()+= maxlin;
		lin ()+= maxlin;
		if (lin () >= text.lines () )
			lin ()= text.lines () - 1;
	}
	else
	{
		inilin ()= text.lines () - 1;
		lin ()= inilin ();
	}
	adjustmaxcol ();
	if (fSelecting)
		movesel (opt, lin (), text [lin () ].col2pos (col () ) );
	//putall ();
	redraw_all= true;
}

void Screen::inicio ()
{
	#if 0
	if (inilin () != 0)
		redraw_all= true;
	inilin () = 0;
	if (inicol () != Col () )
		redraw_all= true;
	inicol ()= Col ();
	lin ()= 0;
	col ()= Col ();
	//putall ();
	//redraw_all= true;
	#else
	gotoline (0);
	#endif
}

void Screen::fin ()
{
	size_t l= ptext->lines () - 1;
	gotoline (l, (* ptext) [l].colmax () );
}

bool Screen::back ()
{
	Text & text= * ptext;
	const TextLine & s= text [lin () ];
	size_t pos= s.col2pos (col () );
	PosText opt;
	if (fSelecting)
		opt.set (lin (), pos);
	if (pos > 0)
	{
		--pos;
		col ()= s.pos2col (pos);
	}
	else
	{
		if (lin () == 0) return false;
		--lin ();
		col ()= text [lin () ].colmax ();
		pos= text [lin ()].size ();
	}
	if (fSelecting)
		movesel (opt, lin (), pos);
	if (lin () < inilin () )
		scrollup ();
	else
		if (fSelecting && lin () != opt.lin)
			putline (opt);
	if (fSelecting)
		putactualline ();
	return true;
}

void Screen::move_left ()
{
	back ();
}

bool Screen::forward ()
{
	Text & text= * ptext;
	const TextLine & s= text [lin () ];
	size_t pos= s.col2pos (col () );
	PosText opt;
	if (fSelecting)
		opt.set (lin (), pos);
	if (pos < s.size () )
	{
		++pos;
		col ()= s.pos2col (pos);
	}
	else
	{
		if (lin () + 1 >= text.lines () )
			return false;
		++lin ();
		col ()= Col ();
		pos= 0;
	}
	if (fSelecting)
		movesel (opt, lin (), pos);
	if (lin () - inilin () > maxlin)
		scrolldown ();
	else
		if (fSelecting && lin () != opt.lin)
			putline (opt);
	if (fSelecting)
		putactualline ();
	return true;
}

void Screen::move_right ()
{
	forward ();
}

void Screen::deletechar ()
{
	const TextLine & s= (* ptext) [lin () ];
	size_t pos= s.col2pos (col () );

	if (pos >= s.size () )
	{
		if (lin () < ptext->lines () - 1)
		{
			ptext->joinline (lin () );
			ot << clrscr;
			putall ();
		}
	}
	else
	{
		ptext->deletecharpos (lin (), pos);
		putactualline ();
	}
}

void Screen::backdelete ()
{
	if (back () )
		deletechar ();
}

void Screen::inschar (unsigned char c)
{
	//TraceFunc tr ("Screen::inschar");

	ptext->insertchar (lin (), col (), c);
}

void Screen::sendchar (unsigned char c)
{
	TraceFunc tr ("Screen::sendchar");

	switch (c)
	{
	case '\n':
		{
			const TextLine & tline= (* ptext) [lin () ];
			size_t pos= tline.col2pos (col () );
			size_t s= tline.size ();
			bool atend= pos == s;
			bool hascr= s > 0 && tline [s - 1] == '\r';
			pos= ptext->insertline
				(lin (), col (), autoindent);
			if (hascr && autocr)
			{
				if (atend)
					ptext->appendchar (lin () + 1, '\r');
				else
					ptext->appendchar (lin (), '\r');
			}
			down ();
			ASSERT (lin () > 0);
			putlineabs (lin () - 1);
			ot << at (lin () - inilin (), 0) << insline;
			putactualline ();
			col ()= (*ptext) [lin () ].pos2col (pos);
		}
		break;
	default:
		inschar (c);
		forward ();
		putactualline ();
	}
}

void Screen::adjustinicol ()
{
	TraceFunc tr ("Screen::adjustinicol");

	if (col () < inicol () )
	{
		tr.message ("adjusting left");
		inicol ()= col ();
		inicol ().previoustab ();
		redraw_all= true;
	}
	else if (col () > inicol () + maxcol)
	{
		tr.message ("adjusting right");
		inicol ()= col () - maxcol;
		inicol ().nexttab ();
		redraw_all= true;
	}
}

void Screen::adjustinilin ()
{
	TraceFunc tr ("Screen::adjustinilin");

	if (lin () < inilin () )
		inilin ()= lin ();
	else
		if (lin () > inilin () + maxlin)
			inilin ()= lin () - maxlin;
		else return;
	redraw_all= true;
}

bool Screen::do_save_as ()
{
	TraceFunc tr ("Screen::do_save_as");

	string strFile;
	int key;
	bool r= dialogGetString (ot,
		key, std::set <int> (),
		"Save as",
		"File name to save", strFile, historyfilename, 60);
	putall ();
	if (! r)
		return false;
	if (strFile.empty () )
		return false;
	ptext->save (strFile);
	return true;
}

bool Screen::do_save ()
{
	TraceFunc tr ("Screen::do_save");

	if (ptext->hasname () )
	{
		ptext->save ();
		return true;
	}
	else
		return do_save_as ();
}

bool Screen::do_save_all ()
{
	TraceFunc tr ("Screen::do_save_all");

	for (size_t i= 0; i < win.size (); ++i)
	{
		Text * aux= win [i].text ();
		if (! aux->saved () )
		{
			if (aux->hasname () )
				aux->save ();
			else
			{
				changewindow (i);
				if (! do_save_as () )
					return false;
			}
		}
	}
	return true;
}

void Screen::do_execute (const std::string & str)
{
	TraceFunc tr ("Screen::do_execute");

	util::bool_saver saveautoindent (autoindent, false);
	status.settext (std::string ("Executing ") + str);
	Command (str);
	status.clear ();
}

bool Screen::question_save ()
{
	TraceFunc tr ("Screen::question_save");

	if (ptext->saved () )
		return true;
	dialogResult dr= dialogQuestion (ot, "Save changes?");
	putall ();
	switch (dr)
	{
	case dialogCancel:
		return false;
	case dialogYes:
		return do_save ();
	default:
		;
	}
	return true;
}

namespace {

class RedrawerScreen : public RedrawerSelectFile
{
public:
	RedrawerScreen (Screen & screen) :
		screen (screen)
	{ }
	void doit ()
	{
		screen.putall ();
	}
private:
	Screen & screen;
};

} // namespace

void Screen::optionLoad ()
{
	TraceFunc tr ("Screen::optionLoad");

	if (! question_save () )
		return;
	string strFile;
	int k;
	std::set <int> keyender;
	keyender.insert (key::F5);
	bool r= dialogGetString (ot,
		k, keyender,
		"Open", "File name to open",
		strFile, historyfilename, 60);
	putall ();
	if (!r && k == key::F5)
	{
		RedrawerScreen redrawer (* this);
		strFile= selectfile (ot, "Load file", redrawer);
		if (! strFile.empty () )
			r= true;
		else
			redraw_all= true;
	}
	if (r)
	{
		loadtext (strFile);
		inittext ();
		redraw_all= true;
	}
}

void Screen::optionSave ()
{
	TraceFunc tr ("Screen::optionSave");

	do_save ();
}

void Screen::optionSaveAs ()
{
	TraceFunc tr ("Screen::optionSaveAs");

	do_save_as ();
}

void Screen::optionQuit ()
{
	TraceFunc tr ("Screen::optionQuit");

	exiting= true;
	for (size_t i= 0, l= win.size (); i < l; ++i)
	{
		if (! closewindow () )
			return;
		putall ();
	}
	fExit= true;
}

void Screen::optionQuitOrClose ()
{
	if (win.size () == 1)
		optionQuit ();
	else
		closewindow ();
}

void Screen::optionCloseWindow ()
{
	TraceFunc tr ("Screen::optionCloseWindow");

	closewindow ();
}

void Screen::optionFileSelect ()
{
	std::vector <std::string> vstr;
	for (size_t i= 0; i < win.size (); ++i)
	{
		Text * pt= win [i].text ();
		if (! pt->hasname () )
			vstr.push_back ("(anonimo)");
		else
			vstr.push_back (pt->name () );
	}
	if (! vstr.empty () )
	{
		size_t r= select (ot, "Seleccionar fichero", vstr);
		if (r != size_t (-1) )
			changewindow (r);
		else
			redraw_all= true;
	}
}

void Screen::optionFileOptions ()
{
	DialogOptions dialog (ot, "File options");
	bool readonly= ptext->readonly ();
	dialog.add ("Solo lectura", readonly);

	if (dialog.do_it () )
	{
		ptext->readonly (readonly);
	}

	putall ();
}

void Screen::optionEditOptions ()
{
	DialogOptions dialog (ot, "Edit options");
	bool optautoindent= autoindent;
	bool optautocr= autocr;
	dialog.add ("Auto indentado", optautoindent);
	dialog.add ("Auto CR", optautocr);

	if (dialog.do_it () )
	{
		autoindent= optautoindent;
		autocr= optautocr;
	}

	putall ();
}

void Screen::optionSelect ()
{
	TraceFunc tr ("Screen::optionSelect");

	if (! fSelecting)
		beginselect ();
	else
		endselect ();
}

void Screen::optionCompile ()
{
	if (! ptext->saved () && ! do_save () )
		return;
	std::string command ("gcc -c ");
	if (! strCompileOptions.empty () )
	{
		command+= strCompileOptions;
		command+= ' ';
	}
	command+= ptext->name ();
	windownew ();
	putall ();
	do_execute (command);
	ptext->modified (false);
	ptext->readonly (true);
}

void Screen::optionMake ()
{
	if (! do_save_all () )
		return;
	windownew ();
	putall ();
	do_execute ("make");
	ptext->modified (false);
	ptext->readonly (true);
}

void Screen::optionCompileOptions ()
{
	TraceFunc tr ("Screen::optionCompileOptions");

	int k;
	bool r= dialogGetString (ot,
		k, std::set <int> (),
		"Compile options",
		"Enter compile options",
		strCompileOptions, historycompileoptions, 60);
	putall ();
	if (r)
		status.settext ("Compile options set to: " +
			strCompileOptions);
}

void Screen::optionPrompt ()
{
	TraceFunc tr ("Screen::optionPrompt");

	bool r= dialogGetString (ot, "Prompt", "Enter new prompt",
		strPrompt, historyprompt, 60);
	putall ();
	if (r)
		status.settext ("Prompt set to: " + strPrompt);
}

void Screen::optionHelp ()
{
	try
	{
		const size_t xmarg= 5, ymarg= 3;
		size_t w= ot.getWidth ();
		size_t x1= xmarg;
		size_t x2= w - xmarg;
		if (x1 + 10 >= x2)
			throw 1;
		w= x2 - x1;
		size_t h= ot.getHeight ();
		size_t y1= ymarg;
		size_t y2= h - ymarg;
		if (y1 + 5 >= y2)
			throw 1;
		h= y2 - y1;
		extern const char strHelp [];
		dialogShowText (ot, "Help", strHelp, h, w);
	}
	catch (...)
	{
		status.settext ("Error showing help");
	}
	putall ();
}

void Screen::optionHelpPerl ()
{
	std::vector <std::string> vstr;
	for (perlfunchelp_t::iterator it= perlfunchelp.begin ();
		it != perlfunchelp.end ();
		++it)
	{
		vstr.push_back (it->first);
	}
	if (vstr.empty () )
		return;
	for (;;)
	{
		size_t r= select (ot, "Seleccionar funcin", vstr);
		putall ();
		if (r == size_t (-1) )
			break;
		dialogShowText (ot, vstr [r].c_str (),
			perlfunchelp [vstr [r] ].c_str (),
			ot.lines () / 2, ot.columns () / 2);
		putall ();
	}
}

void Screen::optionAbout ()
{
	TraceFunc tr ("Screen::about");

	try
	{
		int width= 40;
		int w= ot.getWidth ();
		int x= (w - width) / 2;
		if (x < 0) x= 0;
		if (width > w) width= w;
		size_t h= ot.getHeight ();
		const size_t height= 9;
		if (h < height)
			throw 1;
		size_t y= (h - height) / 2;
		otermstream ots (ot, height, width, y + 1, x);
		ots.cursor (false);
		ots << clrscr;
		ots.border ();
		centertext (ots, 0, width, " qtre ");
		centertext (ots, 2, width, "Quick Tool for Raw Editing");
		centertext (ots, 3, width, "(C) 2000-2003 Julian Albo");
		centertext (ots, 4, width, strVersion);
		centertext (ots, 6, width, "Press any key");
		ots.getkey ();
		ots.cursor (true);
	}
	catch (...) { }
	putall ();
}

bool Screen::askSearchOptions (SearchOptions & options)
{
	std::string strOptions;
	static History historyoptions;
	bool r= dialogGetString (ot, "Options",
		"I-Ignore case, B-Backwards, W-Whole word- "
		"R-Regular expresssion",
		strOptions, historyoptions, 10);
	putall ();
	if (! r)
		return false;
	bool invalid= false;
	for (std::string::size_type i= 0, l= strOptions.size (); i < l; ++i)
	{
		switch (strOptions [i] )
		{
		case 'I': case 'i':
			options.nocase= true;
			break;
		case 'B': case 'b':
			options.backwards= true;
			break;
		case 'W': case 'w':
			options.wholeword= true;
			break;
		case 'R': case 'r':
			options.regex= true;
			break;
		default:
			invalid= true;
		}
		if (invalid)
			throw std::runtime_error ("Invalid option");
	}
	if (options.regex && options.wholeword)
		throw std::runtime_error ("Options R and W are icompatibles");
	return true;
}

void Screen::buildfinder (const SearchOptions & options,
	const std::string & strFind)
{
	Finder * poldfinder= pfinder;

	if (! options.regex)
	{
		if (! options.wholeword)
		{
			if (! options.nocase)
			{
				if (! options.backwards)
					pfinder= new NormalFinder (strFind);
				else
					pfinder= new BackFinder (strFind);
			}
			else
			{
				if (! options.backwards)
					pfinder= new NocaseFinder (strFind);
				else
					pfinder= new
						BackNocaseFinder (strFind);
			}
		}
		else
		{
			if (! options.nocase)
			{
				if (! options.backwards)
					pfinder= new WordFinder (strFind);
				else
					pfinder= new WordBackFinder (strFind);
			}
			else
			{
				if (! options.backwards)
					pfinder= new
						WordNocaseFinder (strFind);
				else
					pfinder= new
						WordBackNocaseFinder (strFind);
			}
		}
	}
	else
	{
		// Testing.
		if (! options.backwards)
			pfinder= new RegexFinder (strFind, options.nocase);
		else
			pfinder= new RegexBackFinder
				(strFind, options.nocase);
	}

	delete poldfinder;
}

void Screen::optionFind ()
{
	TraceFunc tr ("Screen::optionFind");

	bool r= dialogGetString (ot, "Find", "String to find",
		strFind, historyfind, 60);
	putall ();
	if (! r)
	{
		status.clear ();
		return;
	}

	SearchOptions options;
	r= askSearchOptions (options);
	if (! r)
		return;
	lastsearchoptions= options;

	buildfinder (options, strFind);

	optionFindAgain ();
}

#if 0
namespace {

struct equal_nocase {
	bool operator () (const char c1, const char c2) const
	{
		return toupper (c1) == toupper (c2);
	}
};

} // namespace
#endif

bool Screen::searchtext (const std::string & /*searched*/,
	SearchOptions /*options*/)
{
	TraceFunc tr ("Screen::searchtext");

	status.settext ("Searching...");

	Text & text= * ptext;
	// Close possible open selcting.
	//fSelecting= false;
	PosText opt;
	if (fSelecting)
		opt.set (lin (), text [lin () ].col2pos (col () ) );

	//#define OLD_SEARCH

	#ifndef OLD_SEARCH
	ASSERT (pfinder != NULL);
	#endif

	const TextLine & firstline= text [lin () ];
	const string & firststr= firstline.str ();
	string::size_type r;
	bool fSigueBuscando= true;
	string::size_type ini= firstline.col2pos (col () );
	#ifdef OLD_SEARCH
	bool searchfirst;
	if (options.backwards)
	#else
	if (pfinder->backwards () )
	#endif
	{
		#ifdef OLD_SEARCH
		searchfirst= ini > searched.size ();
		if (searchfirst)
			ini-= searched.size () + 1;
		#else
		// Testing, now the ini position is calculated.
		// inside finders.
		//searchfirst= ini > pfinder->length ();
		//if (searchfirst)
		//	ini-= pfinder->length () + 1;
		#endif
	}
	else
	{
		#ifdef OLD_SEARCH
		++ini;
		searchfirst= ini < firststr.size ();
		#else
		// Testing, now the ini position is calculated.
		// inside finders.
		// Then nothing to do here.
		#endif
	}

	#ifdef OLD_SEARCH
	bool alnumfirst= false, alnumlast= false;
	if (options.wholeword && ! searched.empty () )
	{
		alnumfirst= isalnum (searched [0]);
		alnumlast= isalnum (searched [searched.size () - 1] );
	}
	#endif

	size_t l= lin ();

	#ifdef OLD_SEARCH
	while (searchfirst)
	{
		if (! options.nocase)
		{
			if (options.backwards)
				r= firststr.rfind (searched, ini);
			else
				r= firststr.find (searched, ini);
		}
		else
		{
			std::string::const_iterator notfound=
				options.backwards ?
					firststr.begin () + ini :
					firststr.end ();
			std::string::const_iterator it=
				options.backwards ?
					std::find_end (firststr.begin (),
						firststr.begin () + ini,
						searched.begin (),
						searched.end (),
						equal_nocase () )
					:
					std::search (firststr.begin () + ini,
						firststr.end (),
						searched.begin (),
						searched.end (),
						equal_nocase () );
			r= (it == notfound) ? string::npos :
				std::distance (firststr.begin (), it);
		}
		if (r != string::npos)
		{
			if (options.wholeword)
			{
				std::string::size_type end=
					r + searched.size ();
				if ( (! alnumfirst ||
						(r == 0 ||
						! isalnum (firststr [r - 1] ) ) )
					&&
					(! alnumlast ||
						(end >= firststr.size () ||
						! isalnum (firststr [end] )
						) ) )
				{
					fSigueBuscando= false;
					searchfirst= false;
				}
				else
				{
					if (options.backwards)
					{
						//ini= r + searched.size () - 1;
						if (r == 0)
						{
							fSigueBuscando= false;
							searchfirst= false;
						}
						else
							ini= r;
					}
					else
						ini= r + 1;
				}
			}
			else
			{
				fSigueBuscando= false;
				searchfirst= false;
			}
		}
		else
			searchfirst= false;
	}
	#else
	// Testing, now allways call pfinder->find in the first line,
	// the checkings are done inside it.
	//if (searchfirst)
	//{
		r= pfinder->find (firststr, ini);
		fSigueBuscando= r == std::string::npos;
	//}
	#endif

	if (fSigueBuscando)
	{
		//size_t l= lin ();
		//for (l= lin () + 1; l < text.lines (); ++l)
		for (;;)
		{
			#ifdef OLD_SEARCH
			if (options.backwards)
			#else
			if (pfinder->backwards () )
			#endif
			{
				if (l == 0)
					return false;
				--l;
			}
			else
			{
				++l;
				if (l >= text.lines () )
					return false;
			}
			const string & str= text [l].str ();

			#ifdef OLD_SEARCH
			std::string::size_type pos= options.backwards ?
				str.size () : 0;
		find_again:
			if (! options.nocase)
			{
				if (options.backwards)
					r= str.rfind (searched, pos);
				else
					r= str.find (searched, pos);
			}
			else
			{
				std::string::const_iterator notfound=
					options.backwards ?
						str.begin () + pos :
						str.end ();
				std::string::const_iterator it=
					options.backwards ?
						std::find_end (str.begin (),
							//str.end (),
							str.begin () + pos,
							searched.begin (),
							searched.end (),
							equal_nocase () )
						:
						std::search (str.begin () +
								pos,
							str.end (),
							searched.begin (),
							searched.end (),
							equal_nocase () );
				r= (it == notfound) ? string::npos :
					std::distance (str.begin (), it);
			}
			if (r != string::npos)
			{
				if (! options.wholeword)
					break;
				std::string::size_type end=
					r + searched.size ();
				if ( (! alnumfirst || r == 0 ||
						! isalnum (str [r - 1]) )
					&&
					(! alnumlast || end >= str.size () ||
						! isalnum (str [end] ) )
					)
				{
					break;
				}
				if (options.backwards)
				{
					//pos= r + searched.size () - 1;
					if (r == 0)
					{
						continue;
					}
					else
						ini= r;
				}
				else
					pos= r + 1;
				goto find_again;
			}
			#else
			r= pfinder->find (str);
			if (r != std::string::npos)
				break;
			#endif
		}
		if (l >= text.lines () )
			return false;
	}
	ASSERT (r != string::npos);
	#ifdef OLD_SEARCH
	r+= searched.size ();
	#else
	r+= pfinder->length ();
	#endif
	if (l != lin () )
	{
		lin ()= l;
		adjustinilin ();
	}
	col ()= text [lin () ].pos2col (r);

	if (fSelecting)
	{
		movesel (opt, lin (), r);
		redraw_all= true;
	}

	return true;
}

void Screen::optionFindAgain ()
{
	TraceFunc tr ("Screen::optionFindAgain");

	if (! pfinder)
	{
		status.settext ("No previous search");
		return;
	}

	if (searchtext (strFind, lastsearchoptions) )
		status.settext ("Found");
	else
		status.settext ("Not found");
}

void Screen::optionReplace ()
{
	TraceFunc tr ("Screen::optionReplace");

	bool r= dialogGetString (ot, "Replace", "Replace from",
		strFind, historyfind, 60);
	putall ();
	if (! r)
		return;
	r= dialogGetString (ot, "Replace to", "Replace to",
		strReplace, historyreplace, 60);
	putall ();
	if (! r)
		return;

	SearchOptions options;
	r= askSearchOptions (options);
	if (! r)
		return;
	lastsearchoptions= options;

	buildfinder (options, strFind);

	bool ask= true;
	size_t nfound= 0, nreplace= 0;
	for (bool looping= true; looping; )
	{
		if (! searchtext (strFind, options) )
			break;

		++ nfound;
		adjustinicol ();

		Selection & sel= win [actualwin].sel;
		if (sel.begin () != sel.end () )
			redraw_all= true;
		size_t pos= (* ptext) [lin () ].col2pos (col () );
		#ifdef OLD_SEARCH
		sel.init (lin (), pos - strFind.size () );
		#else
		sel.init (lin (), pos - pfinder->length () );
		#endif
		sel.moveend (PosText (lin (), pos) );
		

		if (redraw_all)
			putall ();
		else
			putlineabs (lin () );

		char r= 'Y';
		if (ask)
		{
			r= status.question (
				"Replace? (Yes/No/All/One/Cancel):",
				"YyNnAaOoCc");
		}
		status.settext ("Replacing...");
		bool do_it= false;
		switch (r)
		{
		case 'c': case 'C':
			looping= false;
			break;
		case 'n': case 'N':
			break;
		case 'a': case 'A':
			ask= false;
			do_it= true;
			break;
		case 'y': case 'Y':
			do_it= true;
			break;
		case 'o': case 'O':
			do_it= true;
			looping= false;
		}
		if (do_it)
		{
			++nreplace;
			#ifdef OLD_SEARCH
			std::string::size_type l= strFind.size ();
			#else
			std::string::size_type l= pfinder->length ();
			#endif
			const TextLine & t= (* ptext) [lin () ];
			std::string::size_type pos= t.col2pos (col () ) - l;

			#ifdef OLD_SEARCH
			ptext->replace (lin (), pos, l, strReplace);
			pos+= strReplace.size ();
			#else
			std::string replaced=
				pfinder->replace (strReplace, t.str () );
			ptext->replace (lin (), pos, l, replaced);
			pos+= replaced.size ();
			#endif
			sel.moveend (PosText (lin (), pos) );

			if (looping)
				--pos; // searchtext begins in +1 position
			col ()= t.pos2col (pos);
			adjustinicol ();
			if (redraw_all)
				putall ();
			else
				putlineabs (lin () );
		}
	}
	std::ostringstream oss;
	oss << "Replace: " << nfound << " found, " <<
		nreplace << " replaced";
	status.settext (oss.str () );
}

void Screen::optionGoLine ()
{
	TraceFunc tr ("Screen::optionGoLine");

	string strNum;
	static History history;
	bool r= dialogGetString (ot, "Go to", "Line number",
			strNum, history, 6, 6);
	putall ();
	if (r)
	{
		size_t num= atoi (strNum.c_str () );
		if (num == size_t (-1) || num == 0)
		{
			status.settext ("Invalid number");
			return;
		}
		if (num > ptext->lines () )
		{
			status.settext ("Out of text");
			return;
		}
		gotoline (num - 1);
		status.clear ();
	}
	else
		status.clear ();
}

namespace {

class PopenCloser {
public:
	#if 0
	PopenCloser (FILE * f) : f (f)
	{ }
	#else
	PopenCloser (const std::string str)
	{
		int fnull= open ("/dev/null", O_WRONLY);
		saveerr= dup (STDERR_FILENO);
		dup2 (fnull, STDERR_FILENO);
		close (fnull);
		f= popen (str.c_str (), "r");
	}
	#endif
	~PopenCloser ()
	{
		dup2 (saveerr, STDERR_FILENO);
		if (f)
			pclose (f);
	}
	bool failed () const { return ! f; }
	operator FILE * () { return f; }
private:
	FILE * f;
	int saveerr;
};

} // namespace

void Screen::optionManOf ()
{
	TraceFunc tr ("Screen::optionManOf");

	bool r= dialogGetString (ot, "Man", "Man de",
		strMan, historyman, 60);
	putall ();
	if (! r || strMan.empty () )
		return;
	openman (strMan);
}

void Screen::optionMan ()
{
	TraceFunc tr ("Screen::optionMan");

	std::string str= ptext->wordat (lin (), col () );
	if (str.empty () )
		return;
	openman (str);
}

void Screen::openman (const std::string & str)
{
	TraceFunc tr ("Screen::openman");

	//std::string command= "man " + str + " 2> /dev/null";
	std::string command= "man " + str;
	const std::string st ("executing " + command);
	status.settext (st);
	status.settext (st);
	#if 0
	FILE * f= popen (command.c_str (), "r");
	if (! f)
	{
		status.settext ("Error in popen");
		return;
	}
	PopenCloser pcf (f);
	#else
	PopenCloser f (command);
	if (f.failed () )
	{
		status.settext ("Error in popen");
		return;
	}
	#endif

	// Reads the man page skipping the backspaced
	// chars used for bold and underlined.

	int c;
	int prev= fgetc (f);

	if (prev == EOF)
		status.settext ("No man for " + str);
	else
	{
		windownew ();
		putall ();
		status.settext (st);
		ot << at (0, 0);
		util::bool_saver saveautoindent (autoindent, false);
		util::bool_saver saveautocr (autocr, false);
		util::bool_saver savequietmode (quietmode, true);
	
		while ( (c= fgetc (f) ) != EOF)
		{
			if (c != '\x08')
				sendchar (prev);
			else
				c= fgetc (f);
			prev= c;
		}
		sendchar (prev);
		ptext->modified (false);
		ptext->readonly (true);
		inicio ();
		putall ();
		status.clear ();
	}
}

void Screen::optionShell ()
{
	TraceFunc tr ("Screen::optionShell");

	mainwindow << clrscr;
	otermstream::shellmode ();
	const char * strShell= getenv ("SHELL");
	if (strShell == NULL)
		strShell= "sh";
	tr.message (std::string ("Using shell ") + strShell);
	system (strShell);
	otermstream::progmode ();
	mainwindow << clrscr;
	menu.show (menuwindow);
	putall ();
	status.refresh ();
}

void Screen::optionExecute ()
{
	TraceFunc tr ("Screen::optionExecute");

	string strCommand;
	static History history;
	endselect ();
	bool r= dialogGetString (ot, "Execute", "Command to execute",
			strCommand, history, 60);
	putall ();
	if (! r)
	{
		status.clear ();
		return;
	}
	do_execute (strCommand);
}

void Screen::optionExecuteCurrent ()
{
	TraceFunc tr ("Screen::optionExecuteCurrent");

	string strCommand= (* ptext) [lin () ].str ();
	tr.message (strCommand);
	const std::string::size_type l= strPrompt.size ();
	if (l)
	{
		if (strCommand.substr (0, l) == strPrompt)
			strCommand.erase (0, l);
		else
		{
			home ();
			instext (strPrompt.c_str () );
		}
	}
	end ();
	sendchar ('\n');
	do_execute (strCommand);
	if (l)
		instext (strPrompt.c_str () );
}

void Screen::optionEvaluate ()
{
	TraceFunc tr ("Screen::optionEvaluate");

	static History history;
	bool r= dialogGetString (ot, "Evaluate", "Write a perl expression",
		strEval, history, 60);
	putall ();
	if (r)
	{
		if (qperl_eval (strEval.c_str (), this) != 0)
			status.settext ("Error evaluating");
	}
}

void Screen::optionEvalSelection ()
{
	TraceFunc tr ("Screen::optionEvalSelection");

	Text clipboard;
	copyselection (clipboard);
	std::string str (clipboard.str () );
	if (qperl_eval (str.c_str (), this) != 0)
		status.settext ("Error evaluating selection");
}

void Screen::optionEvalClipboard ()
{
	TraceFunc tr ("Screen::optionEvalClipboard");

	std::string str (pclipboard->str () );
	if (qperl_eval (str.c_str (), this) != 0)
		status.settext ("Error evaluating clipboard");
}

void Screen::optionEvalCurrent ()
{
	TraceFunc tr ("Screen::optionEvalCurrent");

	std::string str= (* ptext) [lin () ].str ();
	end ();
	sendchar ('\n');
	if (qperl_eval (str.c_str (), this) != 0)
		status.settext ("Error evaluating line");
}

void Screen::optionDefineControl ()
{
	TraceFunc tr ("Screen::optionDefineControl");

	std::string key;
	static History historykey;
	bool r= dialogGetString (ot, "Define control key", "Key to define",
			key, historykey, 1, 1);
	putall ();
	if (! r)
		return;
	if (key.empty () )
		return;
	char k= key [0];
	if (! isalpha (k) ||
			! (k >= 'A' && k >= 'Z') || ! (k>= 'a' && k<= 'z') )
		return;
	std::string function;
	static History historydef;
	r= dialogGetString (ot, "Define control key", "Function to execute",
			function, historydef, 60);
	putall ();
	if (! r)
		return;
	definecontrolkey (k, function);
}

void Screen::perlfunc (const std::string & func)
{
	switch (qperl_function (func.c_str (), this) )
	{
	case qpfunction_ok:
		break;
	case qpfunction_notfound:
		status.settext (std::string ("Undefined function ") + func);
		break;
	case qpfunction_failed:
		status.settext (std::string ("Error in function ") + func);
		break;
	}
}

void Screen::macro ()
{
	TraceFunc tr ("Screen::macro");

	int c2= statusgetkey ("Macro key: ");
	if (c2 > 255 || c2 < 0)
		return;

	std::string macro ("macro_");
	// Don't use isalnum because is locale dependant.
	if ( (c2 >= '0' && c2 <= '9') ||
		(c2 >= 'a' && c2 <= 'z') || (c2 >= 'A' && c2 <= 'Z') )
	{
		macro+= char (tolower (c2) );
	}
	else if (c2 >= 1 && c2 <= 'z' - 'a' + 1)
	{
		macro+= "ctrl_";
		macro+= char (c2 - 1 + 'a');
	}
	else
	{
		// Convert to hex the easy way.
		std::ostringstream oss;
		oss.setf (std::ios::uppercase);
		oss << std::hex << c2;
		macro+= oss.str ();
	}
	tr.message (macro);

	perlfunc (macro);
}

void Screen::menuselect ()
{
	TraceFunc tr ("Screen::menuselect");

	ot << cursor (false);
	Action * action= menu.select (menuwindow);
	ot << cursor (true);
	if (action)
		action->doit ();
}

bool Screen::fastkey (int c)
{
	TraceFunc tr ("Screen::fastkey");

	//if (Action * action= mapkeyaction.get (c) )
	if (Action * action= getKeyAction (c) )
	{
		action->doit ();
		return true;
	}
	return false;
}

void Screen::sendkey (int c)
{
	TraceFunc tr ("Screen::sendkey");
	using namespace key;

	if (fastkey (c) )
		return;

	if (c <= 255 && (c >= 32 || c == '\t' || c == '\n') )
	{
		sendchar (c);
	}
}

void Screen::deleteline ()
{
	ptext->deleteline (lin () );
	ot << at (lin () - inilin (), 0) << delline;
	putline (maxlin);
	if (lin () >= ptext->lines () )
		up ();
	else
		adjustmaxcol ();
}

void Screen::replaceline (char *str, int line)
{
	TraceFunc tr ("Screen:replaceline");

	size_t n= static_cast <size_t> (line);
	{
		std::ostringstream oss;
		oss << "Replacing line " << n << " with " << str;
		tr.message (oss.str () );
	}
	ptext->setline (n, str);
	if (n == lin () )
		adjustmaxcol ();
	//putactualline ();
	if (n >= inilin () && n <= inilin () + maxlin)
		if (quietmode)
			redraw_all= true;
		else
			putlineabs (n);
}

void Screen::interact ()
{
	TraceFunc tr ("interact");

	menu.show (menuwindow);
	redraw_all= true;
	int c;
	while (! fExit)
	{
		adjustinicol ();
		if (redraw_all)
			putall ();
		redraw_all= false;
		status.lincol (lin (), col ().getn () );
		ot << at (lin () - inilin (), (col () - inicol () ).getn () );
		tr.message ("Waiting key");
		c= getkey ();
		if (! fExit)
			sendkey (c);
	}
}

void Screen::loadtext (const string & strFileName)
{
	TraceFunc tr ("Screen::loadtext");

	status.settext (string ("Cargando ") + strFileName + "...");
	try
	{
		const char * const argv [1]=
			{ strFileName.c_str () };
		qperl_function_args ("before_load", this,
			util::dim_array (argv), argv);
		if (! ptext->load (strFileName) )
		{
			std::string dir= dirutil::dir_name (strFileName);
			tr.message (dir);
			if (dir.empty () || dirutil::is_directory (dir) )
			{
				tr.message ("New file");
				// Ignore error, the perl function
				// is not mandatory.
				qperl_function ("new_file", this);
				status.settext ("New file");
			}
			else
			{
				tr.message ("Load error");
				dialogShowText (ot, "Error",
					"Directory not exist", 0, 50);
			}
		}
		else
		{
			tr.message ("File loaded");
			status.settext ("File loaded");
			qperl_function_args ("after_load", this,
				util::dim_array (argv), argv);
		}
	}
	catch (std::exception & e)
	{
		tr.message (std::string ("Exception: ") + e.what () );
		dialogShowText (ot, "Error", e.what (), 0, 50);
		status.clear ();
	}
}

void Screen::definecontrolkey (char k, const std::string & function)
{
	if (k >= 'a' && k <= 'z')
		k-= 'a';
	else if (k >= 'A' && k<= 'Z')
		k-= 'A';
	else if (k < key::Ctl_A || k > key::Ctl_Z)
		return;
	// Cast to avoid a warning on gcc 3.3
	ScreenActionPerl * action= new ScreenActionPerl (* this, function);
	//mapkeyaction.add (k + 1, action);
	addKeyAction (k + 1, action);
}

namespace {

struct KeyAction;

class ScreenActionMapKey : public ScreenAction, public MapKeyAction
{
public:
	ScreenActionMapKey (Screen & screen, KeyAction * ka, size_t n);
	void doit ();
};

struct KeyAction {
	int key;
	union {
		void (Screen::*function) ();
		const char * perlfunc;
		KeyAction * pka;
	};
	size_t n;
	enum TypeAction { Function, Perl, Key };
	TypeAction type;
	KeyAction (int key, void (Screen::* function) () ) :
		key (key), function (function), type (Function)
	{ }
	KeyAction (int key, const char * perlfunc) :
		key (key), perlfunc (perlfunc), type (Perl)
	{ }
	KeyAction (int key, KeyAction * pka, size_t n) :
		key (key), pka (pka), n (n), type (Key)
	{ }
	ScreenAction * action (Screen & screen) const
	{
		TraceFunc tr ("KeyAction::action");

		ScreenAction * a;
		switch (type)
		{
		case Function:
			a= new ScreenActionFunction (screen, function);
			break;
		case Perl:
			a= new ScreenActionPerl (screen, perlfunc);
			break;
		case Key:
			a= new ScreenActionMapKey (screen, pka, n);
			break;
		}
		return a;
	}
};

class KeyActionAdder {
public:
	KeyActionAdder (Screen & screen, MapKeyAction & mka) :
		screen (screen),
		mka (mka)
	{ }
	void operator () (const KeyAction & ka)
	{
		mka.addKeyAction (ka.key, ka.action (screen) );
	}
private:
	Screen & screen;
	MapKeyAction & mka;
};

ScreenActionMapKey::ScreenActionMapKey (Screen & screen,
		KeyAction * ka, size_t n) :
	ScreenAction (screen)
{
	std::for_each (ka, ka + n, KeyActionAdder (screen, * this) );
}

void ScreenActionMapKey::doit ()
{
	int key= screen.statusgetkey ("");
	if (key >= 'A' && key <= 'Z')
		key= tolower (key);
	Action * action= getKeyAction (key);
	if (action)
		action->doit ();
}

KeyAction progkeyaction []= {
	KeyAction ('c',            & Screen::optionCompile),
	KeyAction ('m',            & Screen::optionMake),
	KeyAction ('o',            & Screen::optionCompileOptions),
};

KeyAction keyaction []= {
//	KeyAction (key::Home,      & Screen::home),
//	KeyAction (key::End,       & Screen::end),
	KeyAction (key::Home,      "home"),
	KeyAction (key::End,       "end"),
//	KeyAction (key::Up,        & Screen::move_up),
//	KeyAction (key::Down,      & Screen::move_down),
//	KeyAction (key::Left,      & Screen::move_left),
//	KeyAction (key::Right,     & Screen::move_right),
	KeyAction (key::Up,        "up"),
	KeyAction (key::Down,      "down"),
	KeyAction (key::Left,      "left"),
	KeyAction (key::Right,     "right"),
//	KeyAction (key::PageUp,    & Screen::pageup),
//	KeyAction (key::PageDown,  & Screen::pagedown),
	KeyAction (key::PageUp,    "pageup"),
	KeyAction (key::PageDown,  "pagedown"),
//	KeyAction (key::Delete,    & Screen::deletechar),
	KeyAction (key::Delete,    "delete_char"),
//	KeyAction (key::Backspace, & Screen::backdelete),
	KeyAction (key::Backspace, "backspace"),
//	KeyAction (0x7F,           & Screen::backdelete),
	KeyAction (0x7F,           "backspace"),
//	KeyAction (key::Select,    & Screen::optionSelect),
	KeyAction (key::Select,    "toggle_select"),

	KeyAction (key::Ctl_A,     & Screen::optionFindAgain),
//	KeyAction (key::Ctl_C,     & Screen::copy),
	KeyAction (key::Ctl_C,     "copy"),
	KeyAction (key::Ctl_E,     & Screen::optionEvaluate),
	KeyAction (key::Ctl_F,     & Screen::optionFind),
//	KeyAction (key::Ctl_H,     & Screen::backdelete),
	KeyAction (key::Ctl_H,     "backspace"),
//	KeyAction (key::Ctl_K,     & Screen::macro),
	KeyAction (key::Ctl_K,     "macro"),
	KeyAction (key::Ctl_N,     & Screen::menuselect),
	KeyAction (key::Ctl_O,     progkeyaction, dim_array (progkeyaction) ),
//	KeyAction (key::Ctl_Q,     & Screen::optionQuit),
	KeyAction (key::Ctl_Q,     "file_quit"),
	KeyAction (key::Ctl_R,     & Screen::optionReplace),
	//KeyAction (key::Ctl_S,     & Screen::optionSelect),
	KeyAction (key::Ctl_S,     "toggle_select"),
//	KeyAction (key::Ctl_V,     & Screen::paste),
	KeyAction (key::Ctl_V,     "paste"),
//	KeyAction (key::Ctl_X,     & Screen::cut),
	KeyAction (key::Ctl_X,     "cut"),
	//KeyAction (key::Ctl_Y,     & Screen::deleteline),
	KeyAction (key::Ctl_Y,     "delete_line"),

	KeyAction (key::F1,        & Screen::optionHelp),
//	KeyAction (key::F2,        & Screen::menuselect),
//	KeyAction (key::F3,        & Screen::optionSave),
	KeyAction (key::F2,        "file_save"),
	KeyAction (key::F3,        & Screen::optionMan),
//	KeyAction (key::F5,        & Screen::nextwindow),
	KeyAction (key::F5,        "file_next"),
	KeyAction (key::F7,        & Screen::optionMake),
	KeyAction (key::F8,        & Screen::optionQuitOrClose),
	KeyAction (key::F9,        & Screen::optionExecuteCurrent),
	KeyAction (key::F10,       & Screen::menuselect),
};

} // namespace

void Screen::init_actions ()
{
	TraceFunc tr ("Screen::init_actions");

	std::for_each (keyaction, keyaction + dim_array (keyaction),
		KeyActionAdder (* this, * this) );
}

namespace {

struct MenuOption {
	const char * str;
	union {
		void (Screen::*function) ();
		const char * perlfunc;
	};
	enum TypeOption { Function, Perl };
	TypeOption type;
	MenuOption (const char * str, void (Screen::* function) () ) :
		str (str), function (function), type (Function)
	{ }
	MenuOption (const char * str, const char * perlfunc) :
		str (str), perlfunc (perlfunc), type (Perl)
	{ }
	Option * option (Screen & screen) const
	{
		ScreenAction * action;
		switch (type)
		
		{
		case Function:
			action= new ScreenActionFunction (screen, function);
			break;
		case Perl:
			action= new ScreenActionPerl (screen, perlfunc);
			break;
		}
		return new Option (str, action, MenuOwner);
	}
};

MenuOption
mopFile []= {
//	MenuOption ("&Abrir",                 & Screen::optionLoad),
	MenuOption ("&Abrir",                 "file_load"),
	MenuOption ("&Nuevo",                 "file_new"),
//	MenuOption ("&Guardar",               & Screen::optionSave),
	MenuOption ("&Guardar",               "file_save"),
//	MenuOption ("guardar &Como",          & Screen::optionSaveAs),
	MenuOption ("gu&Ardar como",          "file_save_as"),
	MenuOption ("&Siguiente",             "file_next"),
	MenuOption ("&Cerrar",                "file_close"),
//	MenuOption ("&Terminar",              & Screen::optionQuit),
	MenuOption ("&Terminar",              "file_quit"),
	MenuOption ("s&Eleccionar",           & Screen::optionFileSelect),
	MenuOption ("&Opciones",              & Screen::optionFileOptions),
},
#if 0
mopWindow []= {
//	MenuOption ("&Nueva",                 & Screen::windownew),
	MenuOption ("&Nueva",                 "window_new"),
//	MenuOption ("&Siguiente",             & Screen::nextwindow),
	MenuOption ("&Siguiente",             "window_next"),
//	MenuOption ("&Cerrar",                & Screen::optionCloseWindow),
	MenuOption ("&Cerrar",                "window_close"),
},
#endif
mopEdit []= {
//	MenuOption ("&Seleccionar",           & Screen::optionSelect),
//	MenuOption ("&Cortar",                & Screen::cut),
//	MenuOption ("c&Opiar",                & Screen::copy),
//	MenuOption ("&Pegar",                 & Screen::paste),
	MenuOption ("&Seleccionar",           "toggle_select"),
	MenuOption ("cor&Tar",                "cut"),
	MenuOption ("&Copiar",                "copy"),
	MenuOption ("&Pegar",                 "paste"),
	MenuOption ("b&Orrar",                & Screen::deleteselection),
	MenuOption ("sel. &Todo",             "select_all"),
	MenuOption ("opc&Iones",              & Screen::optionEditOptions),
},
mopFind []= {
	MenuOption ("&Buscar",                & Screen::optionFind),
	MenuOption ("buscar de &Nuevo",       & Screen::optionFindAgain),
	MenuOption ("&Reemplazar",            & Screen::optionReplace),
},
mopGo []= {
	MenuOption ("&Inicio doc.",           & Screen::inicio),
	MenuOption ("&Final doc.",            & Screen::fin),
	MenuOption ("linea &Numero",          & Screen::optionGoLine),
},
mopSystem []= {
	MenuOption ("&Shell",                 & Screen::optionShell),
	MenuOption ("&Ejecutar",              & Screen::optionExecute),
	MenuOption ("ejecutar &Linea",        & Screen::optionExecuteCurrent),
	MenuOption ("&Prompt",                & Screen::optionPrompt),
},
mopPerl []= {
	MenuOption ("e&Valuar",               & Screen::optionEvaluate),
	MenuOption ("evaluar sele&Ccion",     & Screen::optionEvalSelection),
	MenuOption ("evaluar &Portatpapeles", & Screen::optionEvalClipboard),
	MenuOption ("evaluar &Linea",         & Screen::optionEvalCurrent),
	MenuOption ("&Define control",        & Screen::optionDefineControl),
},
mopProgram []= {
	MenuOption ("&Compile",               & Screen::optionCompile),
	MenuOption ("&Make",                  & Screen::optionMake),
	MenuOption ("&Options",               & Screen::optionCompileOptions),
	MenuOption ("m&An",                   & Screen::optionMan),
	MenuOption ("man &De",                & Screen::optionManOf),
},
mopHelp []= {
	MenuOption ("&Indice",                & Screen::optionHelp),
	MenuOption ("funciones &Perl",        & Screen::optionHelpPerl),
	MenuOption ("&Acerca de",             & Screen::optionAbout),
//},
//mopTest []= {
//	MenuOption ("pruebecilla",            "pruebecilla"),
};

class MenuOptionAdder {
public:
	MenuOptionAdder (Screen & screen, Menu & menu) :
		screen (screen), menu (menu)
	{ }
	void operator () (const MenuOption & mop)
	{
		menu.addOption (mop.option (screen), MenuOwner);
	}
private:
	Screen & screen;
	Menu & menu;
};

Menu * createSub (const MenuOption * mop, size_t nopts, Screen & screen)
{
	TraceFunc tr ("createSub");

	Menu * pmenu= new Menu;
	std::for_each (mop, mop + nopts, MenuOptionAdder (screen, * pmenu) );
	return pmenu;
}

struct MenuList {
	const char * str;
	const MenuOption * mop;
	size_t nops;
	Option * option (Screen & screen) const
	{
		Menu * psubmenu= createSub (mop, nops, screen);
		return new Option (str, psubmenu, MenuOwner);
	}
};

MenuList menulist []= {
	{ "&Archivo",      mopFile,    dim_array (mopFile) },
	//{ "&Ventana",      mopWindow,  dim_array (mopWindow) },
	{ "&Edicion",      mopEdit,    dim_array (mopEdit) },
	{ "&Buscar",       mopFind,    dim_array (mopFind) },
	{ "&Ir",           mopGo,      dim_array (mopGo) },
	{ "&Sistema",      mopSystem,  dim_array (mopSystem) },
	{ "&Perl",         mopPerl,    dim_array (mopPerl) },
	{ "p&Rogramacion", mopProgram, dim_array (mopProgram) },
	{ "a&Yuda",        mopHelp,    dim_array (mopHelp) },
//	{ "prueba",        mopTest,    dim_array (mopTest) },
};

class MenuAdder {
public:
	MenuAdder (Screen & screen, Menu & menu) :
		screen (screen), menu (menu)
	{ }
	void operator () (const MenuList & ml)
	{
		menu.addOption (ml.option (screen), MenuOwner);
	}
private:
	Screen & screen;
	Menu & menu;
};

void createMenu (Menu * pmenu, Screen & screen)
{
	TraceFunc tr ("createMenu");

	std::for_each (menulist, menulist + dim_array (menulist),
		MenuAdder (screen, * pmenu) );
}

} // namespace

int qtre_main (int argc, char * * argv, char * * env)
{
	TraceFunc tr ("qtre_main");

	otermstream mainwindow;
	mainwindow.setcolor (1, Color::Cyan, Color::Black);
	int lin= mainwindow.lines ();
	int col= mainwindow.columns ();

	otermstream otmenubar (1, col, 0, 0);
	StatusLine stat (col, lin - 1);
	otmenubar.color (1);

	auto_ptr <Menu> pmenu (new Menu);

	stat.settext ("Estado");

	otermstream textwindow (lin - 2, col, 1, 0);
	Screen screen (stat, * pmenu, textwindow, mainwindow, otmenubar);
	createMenu (pmenu.get (), screen);

	if (qperl_init (env, & screen) )
		return 1;

	if (argc > 1)
	{
		screen.loadtext (argv [1] );
		for (int i= 2; i < argc; ++ i)
		{
			screen.windownew ();
			screen.loadtext (argv [i] );
		}
		if (argc > 2)
			screen.changewindow (0);
	}

	for (;;)
	{
		try
		{
			screen.interact ();
			break;
		}
		catch (std::exception & e)
		{
			dialogShowText (textwindow, "Error", e.what (), 0, 50);
		}
		catch (const std::string & str)
		{
			dialogShowText
				(textwindow, "Error", str.c_str (), 0, 50);
		}
	}

	qperl_destroy ();
	return 0;
}

void handleSigTerm (int)
{
	fExit= true;
}

int main (int argc, char * * argv, char * * env)
{
	TraceFunc tr ("main");
	using std::cout;
	using std::endl;

	signal (SIGUSR1, TraceFunc::show);
	#if 0
	struct sigaction action;
	action.sa_handler= handleSigTerm;
	action.sa_flags= SA_RESTART;
	sigaction (SIGTERM, & action, NULL);
	#endif
	int r= 0;
	try
	{
		r= qtre_main (argc, argv, env);
	}
	catch (std::exception &e)
	{
		cout << endl << "Error: " << e.what ();
		r= 127;
	}
	catch (...)
	{
		cout << endl << "Unexpected error";
		r= 127;
	}

	cout << endl;
	return r;
} // main

// End of qtre.cpp
