# qtre.pl
# Global macros and initializations for qtre.
# Revision 29-nov-2003

package qtre;

use strict;

#*********************************************************************
#	Define keys
#*********************************************************************

#define_control 'C', 'copy';
#define_control 'H', 'backspace';
#define_control 'L', 'redraw';
#define_control 'Q', 'quit';
#define_control 'S', 'toggle_select';
define_control 'T', 'insert_special_char';
#define_control 'V', 'paste';
#define_control 'X', 'cut';
#define_control 'Y', 'delete_line';

sub insert_special_char ()
{
	my $c= get_key;
	$c= 13 if ($c == 10);
	insert_char $c if ($c >= 0 && $c < 256);
}

#*********************************************************************
#	Redirect perl standard output to text insertion.
#*********************************************************************

{
	package QtreRedirectOutput;

	sub TIEHANDLE { bless \( my $scalar ), shift; }
	sub PRINT
	{
		my $self= shift;
		qtre::insert (join "", @_);
	}
	sub PRINTF
	{
		my $self= shift;
		my $fmt= shift;
		qtre::insert (sprintf $fmt, @_);
	}
}

tie ( *QTREOUTPUT, 'QtreRedirectOutput') or die "Failed!\n";

select QTREOUTPUT;



#*********************************************************************
#	Auxiliary functions
#*********************************************************************

sub date
{
	my ($day, $mon, $year)= (localtime) [3, 4, 5];
	$year+= 1900;
	my $m= ('jan', 'feb', 'mar', 'apr', 'may', 'jun',
		'jul', 'aug', 'sep', 'oct', 'nov', 'dec') [$mon];
	$day . '-' . $m . '-' . $year;
}

#*********************************************************************
#	Events
#*********************************************************************

sub new_file
{
	my $ext= get_fileext;
	my $name= get_filename;
	if ($ext eq "cpp" || $ext eq "h")
	{
		insert ("// " . $name . "\n// Revision " . date .
			"\n\n\n\n// End of $name\n" );
		goto_line 3;
	}
	if ($ext eq "c")
	{
		insert ("/*\n\t$name \n\tRevision " . date .
			"\n*/\n\n\n\n/* End of $name */\n" );
		goto_line 5;
	}
	modified 0;
}

#*********************************************************************
#	Macros
#*********************************************************************

#
#	Replace space indenting with tabs in the selection.
#

sub macro_ctrl_i
{
	my ($begline, $begpos)= get_beginsel;
	return unless defined $begline;
	my ($endline, $endpos)= get_endsel;
	++$begline if $begpos;
	--$endline if ! $endpos;
	my $count= 0;
	my $i;
	for ($i= $begline; $i <= $endline; ++$i)
	{
		#status $i;
		my $l= get_line $i;
		my $n= 0;
		my $pos= 0;
		while (substr ($l, $pos, 8) eq '        ')
			{ ++$n; $pos+= 8; }
		if ($n > 0)
		{
			$l= "\t" x $n . substr ($l, $pos);
			replace_line ($l, $i);
			++ $count;
		}
	}
	status 'Lines tabbed: ' . $count;
}

#
# Find the matching character.
#

sub macro_28 # Macro (
{
	my $c= get_char ();
	$_= $c;
	my $forward= /[\(\[{"']/;
	return if (! tr/([{)]}"'/)]}([{"'/ );
	my $s= $_;
	$c= "" if ($c eq '"' || $c eq "'");
	my $endline= num_lines () - 1;
	my $deep= 1;
	mainloop: while (1)
	{
		if ($forward)
		{
			last mainloop if
				(line () == $endline &&
				get_pos () == char_line () );
			right;
		}
		else
		{
			last mainloop if
				(line () == 0 && get_pos () == 0);
			left;
		}
		my $actual= get_char ();
		if ($actual eq $c) { ++$deep; }
		else
		{
			if ($actual eq $s)
			{
			      --$deep;
			      last mainloop if ($deep == 0);
			}
		}
	}
}

#
#	Comment selection
#

sub macro_c
{
	my ($begline, $begpos)= get_beginsel;
	return unless defined $begline;
	end_select;
	my ($endline, $endpos)= get_endsel;
	++$begline if $begpos;
	--$endline if ! $endpos;
	my $ext= get_fileext;
	my $comment;
	if ($ext eq 'cpp' || $ext eq 'h')
		{ $comment= '//'; }
	else
		{ $comment= '#'; }

	my $i;
	for ($i= $begline; $i <= $endline; ++$i)
	{
		replace_line $comment . get_line ($i), $i
	}
}

#
#	Indent selection
#

sub macro_i
{
	my ($begline, $begpos)= get_beginsel;
	return unless defined $begline;
	end_select;
	my ($endline, $endpos)= get_endsel;
	++$begline if $begpos;
	--$endline if ! $endpos;
	my $i;
	my $tab= "\t";
	for ($i= $begline; $i <= $endline; ++$i)
	{
		replace_line $tab . get_line ($i), $i
	}
}

#
#	Deletes the ^M at end of line of ms-dos files
#

sub macro_m ()
{
	my $fin= num_lines ();
	my $i= 0;
	my $t;
	while ($i < $ fin)
	{
		$t= get_line $i;
		if (substr ($t, length ($t) - 1, 1) eq "\r")
		{
			chop ($t);
			replace_line $t, $i;
		}
		++$i;
	}
}

#
#	Uncomment selection
#

sub macro_t
{
	my ($begline, $begpos)= get_beginsel;
	return unless defined $begline;
	end_select;
	my ($endline, $endpos)= get_endsel;
	++$begline if $begpos;
	--$endline if ! $endpos;
	my $ext= get_fileext;
	my $comment;
	if ($ext eq 'cpp' || $ext eq 'h')
		{ $comment= '//'; }
	else
		{ $comment= '#'; }
	my $lcom= length $comment;

	my $i;
	for ($i= $begline; $i <= $endline; ++$i)
	{
		my $l= get_line $i;
		if (substr ($l, 0, $lcom) eq $comment)
		{
			replace_line substr ($l, $lcom), $i
		}
	}
}

#
#	Unindent selection
#

sub macro_u
{
	my ($begline, $begpos)= get_beginsel;
	return unless defined $begline;
	end_select;
	my ($endline, $endpos)= get_endsel;
	++$begline if $begpos;
	--$endline if ! $endpos;
	my $i;
	my $tab= "\t";
	my $spaces8= ' ' x 8;
	for ($i= $begline; $i <= $endline; ++$i)
	{
		my $l= get_line $i;
		if (substr ($l, 0, 1) eq $tab)
		{
			replace_line substr ($l, 1), $i;
		}
		else
		{
			if (substr ($l, 0, 8) eq $spaces8)
			{
				replace_line substr ($l, 8), $i;
			}
		}
	}
}

#*********************************************************************

1;

# End of qtre.pl
