// termstream.cpp
// Revision 26-nov-2003

#include "termstream.h"

#include <ncurses.h>

#undef border

int termbuf::Control::n= 0;

termbuf::Control::Control ()
{
	if (n == 0)
	{
		initscr ();
		noecho ();
		raw ();
		//keypad (stdscr, TRUE); // It's done in setWindow
		start_color ();
	}
	++n;
}

termbuf::Control::~Control ()
{
	if (--n == 0)
		endwin ();
}

otermstream::otermstream (std::ios::openmode which) :
	termstreambase (which),
	std::ostream (& __my_tb),
	isStd (true)
{
	setWindow (stdscr);
}

otermstream::otermstream (int nlines, int ncols, int y, int x) :
	termstreambase (std::ios::out),
	std::ostream (& __my_tb),
	isStd (false)
{
	WINDOW * sw= newwin (nlines, ncols, y, x);
	if (! sw)
		throw "Al diablo";
	setWindow (sw);
}

otermstream::otermstream (otermstream & ot, int nlines, int ncols, int y, int x) :
	termstreambase (std::ios::out),
	std::ostream (& __my_tb),
	isStd (false)
{
	//WINDOW * sw= subwin ( (WINDOW *) ot.w, nlines, ncols, y, x);
	WINDOW * sw= derwin ( (WINDOW *) ot.w, nlines, ncols, y, x);
	if (! sw)
		throw "Al diablo";
	setWindow (sw);
}

otermstream::~otermstream ()
{
	if (! isStd)
		delwin ( (WINDOW *) w);
}

void otermstream::update ()
{
	//wnoutrefresh ( (WINDOW *) w);
	//doupdate ();
	wrefresh ( (WINDOW *) w);
}

void otermstream::border ()
{
	wborder ( (WINDOW *) w, 0, 0, 0, 0, 0, 0, 0, 0);
}

void otermstream::clrscr ()
{
	sync ();
	wclear ( (WINDOW *) w);
	wrefresh ( (WINDOW *) w);
}

void otermstream::clreol ()
{
	sync ();
	wclrtoeol ( (WINDOW *) w);
	wrefresh ( (WINDOW *) w);
}

void otermstream::insertline ()
{
	sync ();
	winsertln ( (WINDOW *) w);
}

void otermstream::deleteline ()
{
	sync ();
	wdeleteln ( (WINDOW *) w);
}

void otermstream::at (int y, int x)
{
	sync ();
	wmove ( (WINDOW *) w, y, x);
}

void otermstream::inverse (bool flag)
{
	sync ();
	if (flag) wattron ( (WINDOW *) w, A_REVERSE);
	else wattroff ( (WINDOW *) w, A_REVERSE);
}

void otermstream::delay (bool flag)
{
	nodelay ( (WINDOW *) w, flag ? false : true);
}

void otermstream::cursor (bool flag)
{
	curs_set (flag ? 1 : 0);
}

int otermstream::getBegX ()
{
	WINDOW * win= (WINDOW *) w;
	int y, x;
	getbegyx (win, y, x);
	return x;
}

int otermstream::getBegY ()
{
	WINDOW * win= (WINDOW *) w;
	int y, x;
	getbegyx (win, y, x);
	return y;
}

int otermstream::getWidth ()
{
	WINDOW * win= (WINDOW *) w;
	int y, x;
	getmaxyx (win, y, x);
	return x;
}

int otermstream::getHeight ()
{
	WINDOW * win= (WINDOW *) w;
	int y, x;
	getmaxyx (win, y, x);
	return y;
}

namespace {

short color_curses (Color::value c)
{
	short v;
	switch (c) {
	case Color::Black:   v= COLOR_BLACK;   break;
	case Color::Red:     v= COLOR_RED;     break;
	case Color::Green:   v= COLOR_GREEN;   break;
	case Color::Yellow:  v= COLOR_YELLOW;  break;
	case Color::Blue:    v= COLOR_BLUE;    break;
	case Color::Magenta: v= COLOR_MAGENTA; break;
	case Color::Cyan:    v= COLOR_CYAN;    break;
	case Color::White:   v= COLOR_WHITE;   break;
	default:             v= COLOR_WHITE;    break;
	}
	return v;
}

} // namespace

void otermstream::setcolor (short n, Color::value f, Color::value b)
{
	init_pair (n, color_curses (f), color_curses (b) );
}

void otermstream::color (short c)
{
	sync ();
	wattrset ( (WINDOW *) w, COLOR_PAIR (c) );
}

void otermstream::background (short c)
{
	sync ();
	wbkgd ( (WINDOW *) w, COLOR_PAIR (c) );
}

void termstreambase::shellmode ()
{
	reset_shell_mode ();
}

void termstreambase::progmode ()
{
	reset_prog_mode ();
}

void termstreambase::setWindow (void *window)
{
	w= window;
	__my_tb.setWindow (w);
	keypad ( (WINDOW *) window, TRUE);
}

int termstreambase::lines ()
{
	int r, nousado;
	getmaxyx ( (WINDOW *) w, r, nousado);
	return r;
}

int termstreambase::columns ()
{
	int r, nousado;
	getmaxyx ( (WINDOW *) w, nousado, r);
	return r;
}

int termbuf::getkey ()
{
	using namespace key;
	sync ();
	int c= wgetch ( (WINDOW *) w);
	if (c & ~ 0xFF)
		switch (c)
		{
		case KEY_LEFT:      c= Left;      break;
		case KEY_RIGHT:     c= Right;     break;
		case KEY_UP:        c= Up;        break;
		case KEY_DOWN:      c= Down;      break;
		case KEY_PPAGE:     c= PageUp;    break;
		case KEY_NPAGE:     c= PageDown;  break;
		case KEY_BACKSPACE: c= Backspace; break;
		case KEY_DC:        c= Delete;    break;
		case KEY_HOME:      c= Home;      break;
		case KEY_FIND:      c= Home;      break;
		case KEY_END:       c= End;       break;
		case KEY_SELECT:    c= Select;    break;
		// Need this assign of KEY_NEXT when TERM is vt320
		case KEY_NEXT:      c= '\t';      break;
		case KEY_F(1):  c= F1;  break;
		case KEY_F(2):  c= F2;  break;
		case KEY_F(3):  c= F3;  break;
		case KEY_F(4):  c= F4;  break;
		case KEY_F(5):  c= F5;  break;
		case KEY_F(6):  c= F6;  break;
		case KEY_F(7):  c= F7;  break;
		case KEY_F(8):  c= F8;  break;
		case KEY_F(9):  c= F9;  break;
		case KEY_F(10): c= F10; break;
		case KEY_F(11): c= F11; break;
		case KEY_F(12): c= F12; break;
		default: c= 0;
		}
	return c;
}

int termbuf::sync ()
{
	std::streamsize n= pptr () - pbase ();
	for (std::streamsize i= 0; i < n; ++i)
		waddch ( (WINDOW *) w, * (unsigned char *) (pbase () + i) );
	wrefresh ( (WINDOW *) w);
	pbump (-n);
	gbump (egptr () - gptr () );
	return 0;
}

int termbuf::overflow (int ch)
{
	sync ();
	if (ch != EOF) {
		waddch ( (WINDOW *) w, ch);
		wrefresh ( (WINDOW *) w);
	}
	return 0;
}

int termbuf::underflow ()
{
	return 0;
}

// Fin de termstream.cpp
