// text.cpp
// Revision 6-dec-2003

#include <string>
#include <fstream>
#include <sstream>
#include <strstream>
#include <vector>
#include <algorithm>
#include <cerrno>
#include <new>
using std::string;
using std::vector;
using std::for_each;

#include "qtreassert.h"

#include "text.h"
#include "trace.h"

const size_t size_max= size_t (0) - 1;
Col Col::colmax= Col (size_max);

//**************************************
//		TextLine
//**************************************

size_t TextLine::col2pos (Col col) const
{
	size_t pos= 0;
	Col i;
	unsigned char c;
	size_t l= size ();
	while (i < col)
	{
		if (pos >= l)
			return size_max;
		c= s [pos];
		if (c == '\t')
			//i= Text::nexttab (i);
			i.nexttab ();
		else
			if (Text::escapable (c) ) i+= 2;
			else ++i;
		++pos;
	}
	return pos;
}

Col TextLine::pos2col (size_t pos) const
{
	Col col;
	unsigned char c;
	size_t l= size ();
	for (size_t i= 0; i < pos; ++i)
	{
		if (i >= l)
			return Col::max ();
		c= s [i];
		if (c == '\t')
			//col= Text::nexttab (col);
			col.nexttab ();
		else
			if (Text::escapable (c) ) col+= 2;
			else ++col;
	}
	return col;
}

Col TextLine::colmax () const
{
	return pos2col (size () );
}

void TextLine::deletechar (Col col)
{
	size_t pos= col2pos (col);
	ASSERT (pos < s.size () );
	s.erase (pos, 1);
}

void TextLine::deletecharpos (size_t pos)
{
	ASSERT (pos < s.size () );
	s.erase (pos, 1);
}

void TextLine::insertchar (Col col, char c)
{
	size_t pos= col2pos (col);
	if (s.size () <= pos)
	{
		if (s.size () < pos)
			// COMPROBAR: el valor de pos no es adecuado para esto.
			s+= string (pos - s.size (), ' ');
		s+= c;
	}
	else s.insert (pos, 1, c);
}

void TextLine::insertcharpos (size_t pos, char c)
{
	if (s.size () <= pos)
	{
		if (s.size () < pos)
			// COMPROBAR: el valor de pos no es adecuado para esto.
			s+= string (pos - s.size (), ' ');
		s+= c;
	}
	else s.insert (pos, 1, c);
}

std::string TextLine::wordat (Col col) const
{
	size_t pos= col2pos (col);
	//if (! isalnum (s [pos] ) )
	//	return std::string ();
	while (pos > 0 && isalnum (s [pos - 1] ) )
		--pos;
	size_t end= pos + 1;
	while (end < s.size () && isalnum (s [end] ) )
		++end;
	return s.substr (pos, end - pos);
}

std::ostream & operator << (std::ostream &os, const TextLine & tl)
{
	os << tl.s;
	return os;
}

//**************************************
//		Text::Error
//**************************************

Text::Error::Error (const string & str) :
	runtime_error (str)
{ }

Text::ErrorSaving::ErrorSaving (const string & filename) :
	Error (string ("Error al guardar ") + filename +
		": " + strerror (errno) )
{ }

Text::ErrorReadOnly::ErrorReadOnly () :
	Error ("Cannot write, read only file")
{ }

//**************************************
//		Text
//**************************************

namespace {

bool do_load (const string & filename, Text::VecLine & t);
bool set_text (const char * str, Text::VecLine & t);
void do_save (const string & filename, const Text::VecLine & t);

} // namespace

Text::Text () :
	isSaved (true),
	isReadOnly (false)
{
	ASSERT (t.empty () );
	t.push_back (TextLine () );
	nlines= 1;
}

void Text::checkreadonly ()
{
	if (isReadOnly)
		throw ErrorReadOnly ();
}

std::string Text::str () const
{
	std::string aux;
	for (size_t i= 0; i < nlines; ++i)
	{
		aux+= t [i].str ();
		if (i != nlines - 1)
			aux+= '\n';
	}
	return aux;
}

void Text::c_str (char * str) const
{
	for (size_t i= 0; i < nlines; ++i)
	{
		const string & strline= t [i].str ();
		std::string::size_type s= strline.size ();
		strline.copy (str, s);
		str+= s;
		if (i < nlines - 1)
			* str++= '\n';
	}
	*str= '\0';
}

bool Text::copy (char * & str, size_t & len)
{
	len= size ();
	str= new (std::nothrow) char [len + 1];
	if (! str)
		return false;
	c_str (str);
	return true;
}

void Text::clear ()
{
	TraceFunc tr ("Text::clear");

	checkreadonly ();
	t.clear ();
	t.push_back (TextLine () );
	nlines= 1;
	isSaved= true;
	//filename= string ();
	filename.erase ();
}

bool Text::load (const std::string & strFile)
{
	TraceFunc tr ("Text::load");

	checkreadonly ();
	bool r= do_load (strFile, t);
	nlines= t.size ();
	filename= strFile;
	isSaved= true;
	return r;
}

bool Text::set (const char * str)
{
	checkreadonly ();
	bool r= set_text (str, t);
	nlines= t.size ();
	filename.erase ();
	isSaved= false;
	return r;
}

void Text::save ()
{
	if (filename.size () == 0)
		throw Error ("No puedo guardar fichero sin nombre");
	do_save (filename, t);
	isSaved= true;
}

void Text::save (const std::string & strFile)
{
	do_save (strFile, t);
	filename= strFile;
	isReadOnly= false;
	isSaved= true;
}

size_t Text::size () const
{
	size_t l= t.size ();
	size_t r= l - 1;
	for (size_t i= 0; i < l; ++i)
		r+= t [i].size ();
	return r;
}

bool Text::empty () const
{
	return t.size () == 1 && t [0].empty ();
}

void Text::setline (size_t lin, const std::string & str)
{
	ASSERT (lin < t.size () );
	checkreadonly ();
	t [lin]= str;
	isSaved= false;
}

void Text::deletecharpos (size_t lin, size_t pos)
{
	ASSERT (lin < t.size () );
	checkreadonly ();
	t [lin].deletecharpos (pos);
	isSaved= false;
}

void Text::insertchar (size_t lin, Col col, unsigned char c)
{
	ASSERT (lin < t.size () );
	checkreadonly ();
	t [lin].insertchar (col, c);
	isSaved= false;
}

void Text::insertcharpos (size_t lin, size_t pos, unsigned char c)
{
	ASSERT (lin < t.size () );
	checkreadonly ();
	t [lin].insertcharpos (pos, c);
	isSaved= false;
}

void Text::append (size_t lin, const std::string & str)
{
	ASSERT (lin < t.size () );
	checkreadonly ();
	t [lin]+= str;
	isSaved= false;
}

void Text::appendchar (size_t lin, unsigned char c)
{
	ASSERT (lin < t.size () );
	checkreadonly ();
	t [lin]+= c;
	isSaved= false;
}

size_t Text::insertline (size_t lin, Col col, bool autoindent)
{
	checkreadonly ();
	//t.insert (t.begin () + lin + 1);
	// Problem in gcc 3.3, workaround.
	t.insert (t.begin () + lin + 1, TextLine () ); 

	TextLine & ant= t [lin];
	TextLine & sigue= t [lin + 1];
	size_t pos= ant.col2pos (col);
	if (pos < ant.size () )
	{
		sigue= ant.substr (pos);
		ant= ant.substr (0, pos);
	}
	++nlines;
	isSaved= false;
	// Autoindent, testing
	size_t ipos= 0;
	if (autoindent)
	{
		size_t l= ant.size ();
		size_t i= 0;
		Col icol;
		//while (i < l && isspace (ant [i]) && ant [i] != '\r')
		for ( ; i < l && (ant [i] == ' ' || ant [i] == '\t'); ++i)
		{
			if (ant [i] == ' ')
				++icol;
			else
				//icol= (icol / 8 + 1) * 8;
				icol.nexttab ();
		}
		if (i > 0)
		{
			//sigue= ant.substr (0, i) + sigue.str ();
			std::string pre;
			size_t ntab= icol.getn () / 8;
			if (ntab)
			{
				pre= std::string (ntab, '\t');
				ipos= ntab;
			}
			size_t nspace= icol.getn () % 8;
			if (nspace)
			{
				pre+= std::string (nspace, ' ');
				ipos+= nspace;
			}
			sigue.insert (pre);
		}
	}
	return ipos;
}

void Text::insertline (size_t lin, const std::string & str)
{
	checkreadonly ();
	t.insert (t.begin () + lin, TextLine (str) );
	++nlines;
	isSaved= false;
}

void Text::deleteline (size_t lin)
{
	checkreadonly ();
	if (lin < nlines - 1)
	{
		t.erase (t.begin () + lin);
		--nlines;
	}
	else
		t [lin]= "";
	isSaved= false;
}

void Text::joinline (size_t lin)
{
	checkreadonly ();
	t [lin]+= t [lin + 1];
	t.erase (t.begin () + lin + 1);
	--nlines;
	isSaved= false;
}

//**************************************
//		Auxiliares
//**************************************

namespace {

bool load_from_stream (std::istream & in, Text::VecLine & t)
{
	TraceFunc tr ("load_from_stream");

	t.clear ();
	char c;
	string s;
	in.unsetf (std::ios::skipws);
	while (in >> c)
	{
		if (c == '\n')
		{
			t.push_back (s);
			s.erase ();
		}
		else
			s+= c;
	}
	t.push_back (s);
	if (! in.eof () )
		throw std::runtime_error ("Error reading file");
	return true;
}

bool do_load (const string & filename, Text::VecLine & t)
{
	TraceFunc tr ("do_load");
	tr.message (std::string ("loading ") + filename);

	std::ifstream in;
	//in.unsetf (std::ios::skipws);
	in.open (filename.c_str () );
	if (! in.is_open () )
		if (errno != ENOENT)
		{
			std::ostringstream oss;
			oss << "al abrir " << filename << ": (" <<
				errno << ") " << strerror (errno);
			throw std::runtime_error (oss.str () );
		}
		else
		{
			t.clear ();
			t.push_back (TextLine () );
			return false;
		}
	#if 0
	t.clear ();
	char c;
	string s;
	while (in >> c)
	{
		if (c == '\n')
		{
			t.push_back (s);
			s.erase ();
		}
		else
			s+= c;
	}
	t.push_back (s);
	if (! in.eof () )
		throw std::runtime_error ("Error reading file");
	return true;
	#else
	return load_from_stream (in, t);
	#endif
}

bool set_text (const char * str, Text::VecLine & t)
{
	TraceFunc tr ("set_text");

	std::istrstream in (str);
	return load_from_stream (in, t);
}

class savefunc {
public:
	savefunc (std::ostream & nos) : os (nos), flag (false) { }
	void operator () (const TextLine & tl)
	{
		if (flag)
			os << '\n';
		else
			flag= true;
		os << tl;
	}
private:
	std::ostream & os;
	bool flag;
};

void do_save (const string & filename, const Text::VecLine & t)
{
	TraceFunc tr ("do_save");
	tr.message (std::string ("saving ") + filename);

	std::ofstream out (filename.c_str () );
	if (! out.is_open () )
		throw Text::ErrorSaving (filename);
	for_each (t.begin (), t.end (), savefunc (out) );
	out.close ();
	if (! out.good () )
		throw Text::ErrorSaving (filename);
}

} // namespace

// Fin de text.cpp
