// trace.cpp
// Revision 24-sep-2004

#ifndef NDEBUG

#include "trace.h"

#include <iostream>
#include <fstream>
#include <string>
#include <stdexcept>

#include <string.h>
#include <stdlib.h>

using std::cerr;
using std::endl;
using std::ostream;
using std::ofstream;
using std::string;
using std::uncaught_exception;

namespace {

// Stream used to log trace info and messages, if any.
ostream * pout= 0;

// Control of initialization.
bool trace_not_inited= true;

// Indentation level.
size_t indent= 0;

// Linked list of trace objects
// Do not use a standard list to avoid possible dependencies
// of order of intialization.

TraceFunc * initial= NULL;
TraceFunc * * lastpos= & initial;

// Auxiliar functions.

ostream * opentracefile (const char * str)
{
	ofstream * pof=
		new ofstream (str, std::ios::app | std::ios::out);
	if (! pof->is_open () )
	{
		cerr << "Error opening " << str << endl;
		delete pof;
		pof= 0;
	}
	return pof;
}

void showinfo (const char * strenterexit, const char * strfunc)
{
	if (pout)
	{
		* pout << string (indent, ' ') << strenterexit << ' ';

		if (uncaught_exception () )
			* pout << "(throwing) ";

		* pout << strfunc << endl;
	}
}

// Fatal error message

const char BAD_USE []= "Bad use of TraceFunc";

} // namespace

TraceFunc::TraceFunc (const char * strFuncName) :
	strfunc (strFuncName),
	next (NULL)
{
	if (trace_not_inited)
	{
		trace_not_inited= false;
		char * aux= getenv ("TRACEFUNC");
		if (aux)
		{
			if (strcmp (aux, "-") == 0)
				pout= & cerr;
			else
				pout= opentracefile (aux);
		}
	}

	previous= lastpos;
	* lastpos= this;
	lastpos= & next;

	showinfo ("Enter", strfunc);
	++indent;
}

TraceFunc::~TraceFunc ()
{
	--indent;
	showinfo ("Exit", strfunc);

	if (next != NULL)
	{
		cerr << BAD_USE << endl;
		abort ();
	}
	if (lastpos != & next)
	{
		cerr << BAD_USE << endl;
		abort ();
	}
	lastpos= previous;
	if (* lastpos != this)
	{
		cerr << BAD_USE << endl;
		abort ();
	}
	* lastpos= NULL;
}

void TraceFunc::message (const std::string & strMes)
{
	if (pout)
	{
		* pout << string (indent, ' ') << strfunc;

		if (uncaught_exception () )
			* pout << "(throwing) ";

		* pout << ": " << strMes << endl;
	}
}

void TraceFunc::show (int)
{
	// Add a carriage return before line feed to do
	// a clean output even if the terminal is using
	// raw mode.

	cerr << "\r\n";

	if (initial == NULL)
		cerr << "TraceFunc: no calls.";
	else
	{
		cerr << "TraceFunc dump of calls: \r\n";
		for (TraceFunc * act= initial; act != NULL; act= act->next)
			cerr << act->strfunc << "\r\n";
		cerr << "TraceFunc dump ended.";
	}
	cerr << "\r\n";
}

#endif

// End of trace.cpp
